import numpy as np
import ubelt as ub


try:
    from line_profiler import profile
except Exception:
    profile = ub.identity

# TODO: Perhaps dynamically fix the wraped method signatures to agree with
# Boxes when possible. OR just make this efficient (perhaps that is a fools
# errand because single box objects wont be efficient in Python). Probably just
# maintain the wrapper.


class Box:
    """
    Represents a single Box.

    This is a convinience class. For multiple boxes use kwimage.Boxes, which is
    more efficient.

    Currently implemented by storing a Boxes object with one item and indexing
    into it. This implementation could be done more efficiently.

    SeeAlso:
        :class:`kwimage.structs.boxes.Boxes`

    Example:
        >>> from kwimage.structs.single_box import *  # NOQA
        >>> box = Box.random()
        >>> print(f'box={box}')
        >>> #
        >>> box.scale(10).quantize().to_slice()
        >>> #
        >>> sl = (slice(0, 10), slice(0, 30))
        >>> box = Box.from_slice(sl)
        >>> print(f'box={box}')
    """
    __slots__ = ('boxes',)

    def __init__(self, boxes, _check: bool = False):
        # if _check:
        #     raise Exception(
        #         'For now, only construct an instance of this using a class '
        #         ' method, like coerce, from_slice, from_shapely, etc...')
        self.boxes = boxes

    @property
    def format(self):
        return self.boxes.format

    @property
    def data(self):
        return self.boxes.data[0]

    def __nice__(self):
        data_repr = repr(self.data.tolist())
        if '\n' in data_repr:
            data_repr = ub.indent('\n' + data_repr.lstrip('\n'), '    ')
        nice = '{}, {}'.format(self.format, data_repr)
        return nice

    def __str__(self):
        """
        Returns:
            str
        """
        classname = self.__class__.__name__
        nice = self.__nice__()
        return '<{0}({1})>'.format(classname, nice)

    def __repr__(self):
        classname = self.__class__.__name__
        nice = self.__nice__()
        return '<{0}({1})>'.format(classname, nice)

    @classmethod
    def random(self, **kwargs):
        import kwimage
        if kwargs.get('num', 1) != 1:
            raise ValueError('Cannot specify num for Box. Use Boxes instead.')
        kwargs['num'] = 1
        boxes = kwimage.Boxes.random(**kwargs)
        self = Box(boxes, _check=False)
        return self

    @classmethod
    def from_slice(self, slice_, shape=None, clip=True, endpoint=True,
                   wrap=False):
        """
        Example:
            >>> import kwimage
            >>> slice_ = kwimage.Box.random().scale(10).quantize().to_slice()
            >>> new = kwimage.Box.from_slice(slice_)
        """
        import kwimage
        boxes = kwimage.Boxes.from_slice(slice_, shape=shape, clip=clip,
                                         endpoint=endpoint, wrap=wrap)
        self = Box(boxes, _check=False)
        return self

    @classmethod
    def from_shapely(self, geom):
        import kwimage
        boxes = kwimage.Boxes.from_shapely(geom)
        self = Box(boxes, _check=False)
        return self

    @classmethod
    def from_dsize(self, dsize):
        """
        Args:
            dsize (Tuple[int, int]): A width/height tuple.

        Returns:
            Self
        """
        import kwimage
        width, height = dsize
        ltrb = np.array([[0, 0, width, height]])
        boxes = kwimage.Boxes(ltrb, 'ltrb', canonical=True)
        self = Box(boxes, _check=False)
        return self

    @classmethod
    def from_data(self, data, format):
        import kwimage
        boxes = kwimage.Boxes([data], format)
        self = Box(boxes, _check=False)
        return self

    @classmethod
    @profile
    def coerce(cls, data, format=None, **kwargs):
        """
        Create an instance of a box from data.

        Returns:
            Box

        Example:
            >>> import kwimage
            >>> box = kwimage.Box.coerce([1, 2, 3, 4], 'xywh')
            >>> print(f'box = {ub.urepr(box, nl=1)}')
            box = <Box(xywh, [1, 2, 3, 4])>
        """
        if isinstance(data, Box):
            return data
        else:
            import numbers
            import sys
            torch = sys.modules.get('torch', None)
            if isinstance(data, list):
                if data and isinstance(data[0], numbers.Number):
                    data = np.array(data)[None, :]
            if isinstance(data, np.ndarray) or torch and torch.is_tensor(data):
                if len(data.shape) == 1:
                    data = data[None, :]
            # return cls(kwimage.Boxes.coerce(data, **kwargs))
            # inline new coerce code until new version lands
            from kwimage import Boxes
            from shapely.geometry import Polygon
            if isinstance(data, Boxes):
                self = data
            elif isinstance(data, Polygon):
                self = Boxes.from_shapely(data)
            else:
                _arr_data = None
                if isinstance(data, np.ndarray):
                    _arr_data = np.array(data)
                elif isinstance(data, list):
                    _arr_data = np.array(data)

                if _arr_data is not None:
                    if format is None:
                        raise Exception('ambiguous, specify Box format')
                    self = Boxes(_arr_data, format=format)
                else:
                    raise NotImplementedError
            return cls(self)

    @property
    def dsize(self):
        xywh = self.boxes.to_xywh(copy=False)
        width, height = xywh.data[..., 2:4].ravel()
        return (int(width), int(height))

    def translate(self, *args, **kwargs):
        new_boxes = self.boxes.translate(*args, **kwargs)
        new = self.__class__(new_boxes)
        return new

    def warp(self, *args, **kwargs):
        new_boxes = self.boxes.warp(*args, **kwargs)
        new = self.__class__(new_boxes)
        return new

    def scale(self, *args, **kwargs):
        new_boxes = self.boxes.scale(*args, **kwargs)
        new = self.__class__(new_boxes)
        return new

    def clip(self, *args, **kwargs):
        new_boxes = self.boxes.clip(*args, **kwargs)
        new = self.__class__(new_boxes)
        return new

    def quantize(self, *args, **kwargs):
        new_boxes = self.boxes.quantize(*args, **kwargs)
        new = self.__class__(new_boxes)
        return new

    def copy(self, *args, **kwargs):
        new_boxes = self.boxes.copy(*args, **kwargs)
        new = self.__class__(new_boxes)
        return new

    def round(self, *args, **kwargs):
        new_boxes = self.boxes.round(*args, **kwargs)
        new = self.__class__(new_boxes)
        return new

    def pad(self, *args, **kwargs):
        new_boxes = self.boxes.pad(*args, **kwargs)
        new = self.__class__(new_boxes)
        return new

    def resize(self, *args, **kwargs):
        new_boxes = self.boxes.resize(*args, **kwargs)
        new = self.__class__(new_boxes)
        return new

    def intersection(self, other):
        """
        Example:
            >>> import kwimage
            >>> self = kwimage.Box.coerce([0, 0, 10, 10], 'xywh')
            >>> other = kwimage.Box.coerce([3, 3, 10, 10], 'xywh')
            >>> print(str(self.intersection(other)))
            <Box(ltrb, [3, 3, 10, 10])>
        """
        new_boxes = self.boxes.intersection(other.boxes)
        new = self.__class__(new_boxes)
        return new

    def union_hull(self, other):
        """
        Example:
            >>> import kwimage
            >>> self = kwimage.Box.coerce([0, 0, 10, 10], 'xywh')
            >>> other = kwimage.Box.coerce([3, 3, 10, 10], 'xywh')
            >>> print(str(self.union_hull(other)))
            <Box(ltrb, [0, 0, 13, 13])>
        """
        new_boxes = self.boxes.union_hull(other.boxes)
        new = self.__class__(new_boxes)
        return new

    def contains(self, other):
        """
        Examples:
            >>> import kwimage
            >>> self = kwimage.Box.random().scale(10).round()
            >>> other = kwimage.Points.random(10).scale(10).round()
            >>> flags = self.contains(other)
            >>> flags = self.contains(np.array(self.center))
            >>> assert np.all(np.diag(flags))
        """
        flags = self.boxes.contains(other)[0]
        return flags

    def to_ltrb(self, *args, **kwargs):
        """
        Example:
            >>> import kwimage
            >>> self = kwimage.Box.random().to_ltrb()
            >>> assert self.format == 'ltrb'
        """
        return self.__class__(self.boxes.to_ltrb(*args, **kwargs))

    def to_xywh(self, *args, **kwargs):
        """
        Example:
            >>> import kwimage
            >>> self = kwimage.Box.random().to_xywh()
            >>> assert self.format == 'xywh'
        """
        return self.__class__(self.boxes.to_xywh(*args, **kwargs))

    def to_cxywh(self, *args, **kwargs):
        """
        Example:
            >>> import kwimage
            >>> self = kwimage.Box.random().to_cxywh()
            >>> assert self.format == 'cxywh'
        """
        return self.__class__(self.boxes.to_cxywh(*args, **kwargs))

    def toformat(self, *args, **kwargs):
        return self.__class__(self.boxes.toformat(*args, **kwargs))

    def astype(self, *args, **kwargs):
        return self.__class__(self.boxes.astype(*args, **kwargs))

    def corners(self, *args, **kwargs):
        """
        Example:
            >>> import kwimage
            >>> assert kwimage.Box.random().corners().shape == (4, 2)
        """
        return self.boxes.corners(*args, **kwargs)

    def to_boxes(self):
        """
        Returns the underlying "kwimage.Boxes" data structure.
        """
        return self.boxes

    @property
    def aspect_ratio(self):
        """
        Example:
            >>> import kwimage
            >>> assert not ub.iterable(kwimage.Box.random().aspect_ratio)
        """
        return self.boxes.aspect_ratio.ravel()[0]

    @property
    def center(self):
        """
        Example:
            >>> import kwimage
            >>> assert len(kwimage.Box.random().center) == 2
        """
        xs, ys = self.boxes.center
        return xs.ravel()[0], ys.ravel()[0]

    @property
    def center_x(self):
        """
        Example:
            >>> import kwimage
            >>> assert not ub.iterable(kwimage.Box.random().center_x)
        """
        return self.boxes.center_x.ravel()[0]

    @property
    def center_y(self):
        """
        Example:
            >>> import kwimage
            >>> assert not ub.iterable(kwimage.Box.random().center_y)
        """
        return self.boxes.center_y.ravel()[0]

    @property
    def width(self):
        """
        Example:
            >>> import kwimage
            >>> assert not ub.iterable(kwimage.Box.random().width)
        """
        return self.boxes.width.ravel()[0]

    @property
    def height(self):
        """
        Example:
            >>> import kwimage
            >>> assert not ub.iterable(kwimage.Box.random().height)
        """
        return self.boxes.height.ravel()[0]

    @property
    def tl_x(self):
        """
        Example:
            >>> import kwimage
            >>> assert not ub.iterable(kwimage.Box.random().tl_x)
        """
        return self.boxes.tl_x.ravel()[0]

    @property
    def tl_y(self):
        """
        Example:
            >>> import kwimage
            >>> assert not ub.iterable(kwimage.Box.random().tl_y)
        """
        return self.boxes.tl_y.ravel()[0]

    @property
    def br_x(self):
        """
        Example:
            >>> import kwimage
            >>> assert not ub.iterable(kwimage.Box.random().br_y)
        """
        return self.boxes.br_x.ravel()[0]

    @property
    def br_y(self):
        """
        Example:
            >>> import kwimage
            >>> assert not ub.iterable(kwimage.Box.random().br_y)
        """
        return self.boxes.br_y.ravel()[0]

    @property
    def dtype(self):
        return self.boxes.dtype

    @property
    def area(self):
        return self.boxes.area.ravel()[0]

    def to_slice(self, endpoint=True):
        """
        Example:
            >>> import kwimage
            >>> slice_ = kwimage.Box.random(rng=0).scale(10).quantize().to_slice()
            >>> assert slice_ == (slice(5, 8, None), slice(5, 7, None))
        """
        return self.boxes.to_slices(endpoint=endpoint)[0]

    def to_shapely(self):
        """
        Example:
            >>> import kwimage
            >>> kwimage.Box.random().to_shapely()
        """
        return self.boxes.to_shapely()[0]

    def to_polygon(self):
        """
        Example:
            >>> import kwimage
            >>> kwimage.Box.random().to_polygon()
        """
        return self.boxes.to_polygons()[0]

    def to_coco(self):
        """
        Example:
            >>> import kwimage
            >>> kwimage.Box.random().to_coco()
        """
        return list(self.boxes.to_coco())[0]

    def draw_on(self, image=None, color='blue', alpha=None, label=None,
                copy=False, thickness=2, label_loc='top_left'):
        """
        Draws a box directly on an image using OpenCV

        Example:
            >>> # xdoctest: +REQUIRES(module:cv2)
            >>> import kwimage
            >>> self = kwimage.Box.random(scale=256, rng=10, format='ltrb')
            >>> canvas = np.zeros((256, 256, 3), dtype=np.uint8)
            >>> image = self.draw_on(canvas)
            >>> # xdoctest: +REQUIRES(--show)
            >>> # xdoctest: +REQUIRES(module:kwplot)
            >>> import kwplot
            >>> kwplot.figure(fnum=2000, doclf=True)
            >>> kwplot.autompl()
            >>> kwplot.imshow(image)
            >>> kwplot.show_if_requested()
        """
        return self.boxes.draw_on(image=image, color=color, alpha=alpha,
                                  labels=[label], copy=copy,
                                  thickness=thickness, label_loc=label_loc)

    def draw(self, color='blue', alpha=None, label=None, centers=False,
             fill=False, lw=2, ax=None, setlim=False, **kwargs):
        """
        Draws a box directly on an image using OpenCV

        Example:
            >>> # xdoctest: +REQUIRES(module:kwplot)
            >>> import kwimage
            >>> self = kwimage.Box.random(scale=512.0, rng=0, format='ltrb')
            >>> self.translate((-128, -128), inplace=True)
            >>> #image = (np.random.rand(256, 256) * 255).astype(np.uint8)
            >>> # xdoctest: +REQUIRES(--show)
            >>> import kwplot
            >>> kwplot.autompl()
            >>> fig = kwplot.figure(fnum=1, doclf=True)
            >>> #kwplot.imshow(image)
            >>> # xdoctest: +REQUIRES(--show)
            >>> self.draw(color='blue', setlim=1.2)
            >>> # xdoctest: +REQUIRES(--show)
            >>> for o in fig.findobj():  # http://matplotlib.1069221.n5.nabble.com/How-to-turn-off-all-clipping-td1813.html
            >>>     o.set_clip_on(False)
            >>> kwplot.show_if_requested()
        """
        if label is None:
            labels = None
        else:
            labels = [label]
        return self.boxes.draw(color=color, alpha=alpha, labels=labels,
                               centers=centers, fill=fill, lw=lw, ax=ax,
                               setlim=setlim, **kwargs)


def _transfer_docstrings():
    """
    Helper to populate Box docstrings from Boxes. In the future we should may
    want to autogenerate better docstrings statically, but for now lets at
    least add some introspection.

    Ignore:
        from kwimage.structs.single_box import *  # NOQA
    """
    from kwimage.structs import Boxes
    import types

    DYNAMIC_TRANSFER = True

    # First build up what we are going to transfer between.
    corresponding_methods = []
    for name in dir(Box):
        if not name.startswith('_') and hasattr(Boxes, name):
            dst_func = getattr(Box, name)
            if isinstance(dst_func, types.MethodType):
                dst_func = dst_func.__func__
            if isinstance(dst_func, property):
                dst_func = dst_func.fget
            if hasattr(dst_func, '__doc__') and dst_func.__doc__ is None:
                src_func = getattr(Boxes, name)
                if isinstance(src_func, property):
                    src_func = src_func.fget
                if src_func.__doc__ is not None:
                    corresponding_methods.append((name, dst_func, src_func))

    if 0:
        # FIXME: docscrape google should parse block types generically
        from xdoctest.docstr import docscrape_google
        for name, dst_func, src_func in corresponding_methods:
            blocks = docscrape_google.split_google_docblocks(src_func.__doc__)
            new_parts = []
            for block_type, block_value in blocks:
                # Ignore example transfers
                if block_type in {'Example', 'CommandLine', 'Returns', 'TODO'}:
                    # TODO: munge return blocks
                    continue
                indent = '    '
                print(f'block_type={block_type}')
                if block_type == '__DOC__':
                    indent = ''
                else:
                    new_parts.append(block_type + ':')
                new_parts.append(ub.indent(block_value.text, indent))
            new_text = '\n'.join(new_parts)
            print(new_text)

    if DYNAMIC_TRANSFER:
        for name, dst_func, src_func in corresponding_methods:
            dst_func.__doc__ = (
                '\n        This function wraps one of the same name in '
                'kwimage.Boxes, but does not have a docstring of its own. '
                'In the meantime we will show the docstring from Boxes\n'
            ) + src_func.__doc__

if 0:
    _transfer_docstrings()



from nptyping import UInt8
from numpy import ndarray
from typing import Any
from typing import List
from typing import Dict
from nptyping import Shape
from nptyping import Integer
import kwimage
from numpy.typing import ArrayLike
import kwcoco
from typing import Callable
from typing import Tuple
from typing import Sequence
import torch
from nptyping import Bool
from numpy.random import RandomState
import ubelt as ub
from _typeshed import Incomplete
from collections.abc import Generator


class _DetDrawMixin:

    def draw(self,
             color: str = ...,
             alpha: Incomplete | None = ...,
             labels: bool = ...,
             centers: bool = ...,
             lw: int = ...,
             fill: bool = ...,
             ax: Incomplete | None = ...,
             radius: int = ...,
             kpts: bool = ...,
             sseg: bool = ...,
             setlim: bool = ...,
             boxes: bool = ...) -> None:
        ...

    def draw_on(self,
                image: ndarray[Any, UInt8] | None = None,
                color: str | Any | List[Any] = 'blue',
                alpha: float | None = None,
                labels: bool | str | List[str] = True,
                radius: float = 5,
                kpts: bool = True,
                sseg: bool = True,
                boxes: bool = True,
                ssegkw: dict | None = None,
                label_loc: str = 'top_left',
                thickness: int = 2) -> ndarray[Any, UInt8]:
        ...


class _DetAlgoMixin:

    def non_max_supression(
            self,
            thresh: float = 0.0,
            perclass: bool = False,
            impl: str = 'auto',
            daq: bool | Dict = False,
            device_id: Incomplete | None = ...
    ) -> ndarray[Shape['*'], Integer]:
        ...

    def non_max_supress(self,
                        thresh: float = ...,
                        perclass: bool = ...,
                        impl: str = ...,
                        daq: bool = ...):
        ...

    def rasterize(self,
                  bg_size,
                  input_dims,
                  soften: int = ...,
                  tf_data_to_img: Incomplete | None = ...,
                  img_dims: Incomplete | None = ...,
                  exclude=...) -> kwimage.Heatmap:
        ...


class Detections(ub.NiceRepr, _DetAlgoMixin, _DetDrawMixin):
    __datakeys__: Incomplete
    __metakeys__: Incomplete
    data: Dict
    meta: Dict

    def __init__(self,
                 data: Dict[str, ArrayLike] | None = None,
                 meta: Dict[str, object] | None = None,
                 datakeys: List[str] | None = None,
                 metakeys: List[str] | None = None,
                 checks: bool = True,
                 **kwargs) -> None:
        ...

    def __nice__(self):
        ...

    def __len__(self):
        ...

    def copy(self):
        ...

    @classmethod
    def coerce(cls, data: Incomplete | None = ..., **kwargs):
        ...

    @classmethod
    def from_coco_annots(cls,
                         anns: List[Dict],
                         cats: List[Dict] | None = None,
                         classes: kwcoco.CategoryTree | None = None,
                         kp_classes: kwcoco.CategoryTree | None = None,
                         shape: tuple | None = None,
                         dset: kwcoco.CocoDataset | None = None) -> Detections:
        ...

    def to_coco(
            self,
            cname_to_cat: Incomplete | None = ...,
            style: str = 'orig',
            image_id: int | None = None,
            dset: kwcoco.CocoDataset | None = None
    ) -> Generator[dict, None, None]:
        ...

    @property
    def boxes(self):
        ...

    @property
    def class_idxs(self):
        ...

    @property
    def scores(self):
        ...

    @property
    def probs(self):
        ...

    @property
    def weights(self):
        ...

    @property
    def classes(self):
        ...

    def num_boxes(self):
        ...

    def warp(self,
             transform: kwimage.Affine | ndarray | Callable | Any,
             input_dims: Tuple[int, int] | None = None,
             output_dims: Tuple[int, int] | None = None,
             inplace: bool = False) -> Detections:
        ...

    def scale(self,
              factor,
              output_dims: Incomplete | None = ...,
              inplace: bool = ...):
        ...

    def translate(self,
                  offset,
                  output_dims: Incomplete | None = ...,
                  inplace: bool = ...):
        ...

    @classmethod
    def concatenate(cls, dets) -> Detections:
        ...

    def argsort(self, reverse: bool = ...) -> torch.Tensor:
        ...

    def sort(self, reverse: bool = ...) -> kwimage.structs.Detections:
        ...

    def compress(self,
                 flags: ndarray[Any, Bool] | torch.Tensor,
                 axis: int = ...) -> kwimage.structs.Detections:
        ...

    def take(self,
             indices: ndarray[Any, Integer],
             axis: int = ...) -> kwimage.structs.Detections:
        ...

    def __getitem__(self, index):
        ...

    @property
    def device(self):
        ...

    def is_tensor(self):
        ...

    def is_numpy(self):
        ...

    def numpy(self):
        ...

    @property
    def dtype(self):
        ...

    def tensor(self, device=...):
        ...

    @classmethod
    def demo(Detections):
        ...

    @classmethod
    def random(cls,
               num: int = 10,
               scale: float | tuple = 1.0,
               classes: int | Sequence = 3,
               keypoints: bool = False,
               segmentations: bool = False,
               tensor: bool = False,
               rng: int | RandomState | None = None) -> Detections:
        ...


class _UnitDoctTests:
    ...




def test_bbox_isect_failure_case():
    import ubelt as ub
    import numpy as np
    import kwimage
    small_box = kwimage.Boxes(np.array([[119.86356797952962, -0.8880716867719497, 119.86816351547256, -0.892667222714885]]), 'ltrb')
    big_box = kwimage.Boxes(np.array([[119.8631329409319, -0.8610904994264246, 119.92872610984104, -0.9266836683730109]]), 'ltrb')

    print('big_box.height = {!r}'.format(big_box.height))
    print('small_box.height = {!r}'.format(small_box.height))

    # Simple fix to negative height boxes
    small_box = small_box.scale((1, -1))
    big_box = big_box.scale((1, -1))

    isect_box = big_box.intersection(small_box)
    print('isect_box       = {!r}'.format(isect_box))

    isect_area = big_box.isect_area(small_box)
    print('isect_area      = {!r}'.format(isect_area))

    sh_box1 = small_box.to_shapely()[0]
    sh_box2 = big_box.to_shapely()[0]
    sh_isect_box = sh_box1.intersection(sh_box2)
    sh_isect_area = sh_isect_box.area
    print('sh_isect_area   = {!r}'.format(sh_isect_area))

    recon_isect_box = kwimage.Boxes.from_shapely(sh_isect_box)
    print('recon_isect_box = {}'.format(ub.urepr(recon_isect_box, nl=1)))



"""
Numpy implementation of alpha blending based on information in [SO25182421]_
and [WikiAlphaBlend]_.

For more types of blending see [PypiBlendModes]_ and [WikiAlphaBlend]_.

References:
    .. [SO25182421] http://stackoverflow.com/questions/25182421/overlay-numpy-alpha
    .. [WikiAlphaBlend] https://en.wikipedia.org/wiki/Alpha_compositing#Alpha_blending
    .. [WikiBlendModes] https://en.wikipedia.org/wiki/Blend_modes
    .. [PypiBlendModes] https://pypi.org/project/blend-modes/
"""

import numpy as np
from . import im_core


def overlay_alpha_layers(layers, keepalpha=True, dtype=np.float32):
    """
    Stacks a sequences of layers on top of one another. The first item is the
    topmost layer and the last item is the bottommost layer.

    Args:
        layers (Sequence[ndarray]): stack of images
        keepalpha (bool): if False, the alpha channel is removed after blending
        dtype (np.dtype): format for blending computation (defaults to float32)

    Returns:
        ndarray: raster: the blended images

    Example:
        >>> import kwimage
        >>> keys = ['astro', 'carl', 'stars']
        >>> layers = [kwimage.grab_test_image(k, dsize=(100, 100)) for k in keys]
        >>> layers = [kwimage.ensure_alpha_channel(g, alpha=.5) for g in layers]
        >>> stacked = kwimage.overlay_alpha_layers(layers)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> kwplot.imshow(stacked)
        >>> kwplot.show_if_requested()
    """
    layer_iter = iter(layers)
    img1 = next(layer_iter)
    rgb1, alpha1 = _prep_rgb_alpha(img1, dtype=dtype)

    for img2 in layer_iter:
        rgb2, alpha2 = _prep_rgb_alpha(img2, dtype=dtype)
        rgb1, alpha1 = _alpha_blend_inplace(rgb1, alpha1, rgb2, alpha2)

    if keepalpha:
        raster = np.dstack([rgb1, alpha1[..., None]])
    else:
        raster = rgb1
    return raster


def overlay_alpha_images(img1, img2, keepalpha=True, dtype=np.float32,
                         impl='inplace'):
    """
    Places img1 on top of img2 respecting alpha channels.
    Works like the Photoshop layers with opacity.

    Args:
        img1 (ndarray): top image to overlay over img2
        img2 (ndarray): base image to superimpose on
        keepalpha (bool): if False, the alpha channel is removed after blending
        dtype (np.dtype): format for blending computation (defaults to float32)
        impl (str): code specifying the backend implementation

    Returns:
        ndarray: raster: the blended images

    TODO:
        - [ ] Make fast C++ version of this function

    Example:
        >>> import kwimage
        >>> img1 = kwimage.grab_test_image('astro', dsize=(100, 100))
        >>> img2 = kwimage.grab_test_image('carl', dsize=(100, 100))
        >>> img1 = kwimage.ensure_alpha_channel(img1, alpha=.5)
        >>> img3 = kwimage.overlay_alpha_images(img1, img2)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> kwplot.imshow(img3)
        >>> kwplot.show_if_requested()

    Ignore:
        import numpy as np
        import kwimage
        poly = kwimage.Polygon.random().scale((10, 10))

        img2 = np.zeros((10, 10, 4))

        img1 = np.zeros((10, 10))
        indicator = poly.fill(img1)

        to_overlay = np.zeros((10, 10) + (4,), dtype=np.float32)
        to_overlay = kwimage.Mask(indicator, format='c_mask').draw_on(to_overlay, color='lime')
        to_overlay = kwimage.ensure_alpha_channel(to_overlay)
        to_overlay[..., 3] = (indicator > 0).astype(np.float32) * 0.5

        raster = kwimage.overlay_alpha_images(to_overlay, img2)

        # xdoctest: +REQUIRES(--show)
        import kwplot
        kwplot.autompl()
        kwplot.imshow(raster)
        kwplot.show_if_requested()
    """
    rgb1, alpha1 = _prep_rgb_alpha(img1, dtype=dtype)
    rgb2, alpha2 = _prep_rgb_alpha(img2, dtype=dtype)

    # Perform the core alpha blending algorithm
    if impl == 'simple':
        rgb3, alpha3 = _alpha_blend_simple(rgb1, alpha1, rgb2, alpha2)
    elif impl == 'inplace':
        rgb3, alpha3 = _alpha_blend_inplace(rgb1, alpha1, rgb2, alpha2)
    elif impl == 'numexpr1':
        rgb3, alpha3 = _alpha_blend_numexpr1(rgb1, alpha1, rgb2, alpha2)
    elif impl == 'numexpr2':
        rgb3, alpha3 = _alpha_blend_numexpr2(rgb1, alpha1, rgb2, alpha2)
    else:
        raise ValueError('unknown impl={}'.format(impl))

    if keepalpha:
        raster = np.dstack([rgb3, alpha3[..., None]])
        # Note: if we want to output a 255 img we could do something like this
        # out = np.zeros_like(img1)
        # out[..., :3] = rgb3
        # out[..., 3] = alpha3
    else:
        raster = rgb3
    return raster


def _prep_rgb_alpha(img, dtype=np.float32):
    img = im_core.ensure_float01(img, dtype=dtype, copy=False)
    img = im_core.atleast_3channels(img, copy=False)
    c = im_core.num_channels(img)

    if c == 4:
        # rgb = np.ascontiguousarray(img[..., 0:3])
        # alpha = np.ascontiguousarray(img[..., 3])
        rgb = img[..., 0:3]
        alpha = img[..., 3]
    else:
        rgb = img
        alpha = np.ones_like(img[..., 0])
    return rgb, alpha


def _alpha_blend_simple(rgb1, alpha1, rgb2, alpha2):
    """
    Core alpha blending algorithm

    SeeAlso:
        _alpha_blend_inplace - alternative implementation
    """
    c_alpha1 = (1.0 - alpha1)
    alpha3 = alpha1 + alpha2 * c_alpha1

    numer1 = (rgb1 * alpha1[..., None])
    numer2 = (rgb2 * (alpha2 * c_alpha1)[..., None])
    with np.errstate(invalid='ignore'):
        rgb3 = (numer1 + numer2) / alpha3[..., None]
    rgb3[alpha3 == 0] = 0
    return rgb3, alpha3


def _alpha_blend_inplace(rgb1, alpha1, rgb2, alpha2):
    """
    Uglier but faster(? maybe not) version of the core alpha blending algorithm
    using preallocation and in-place computation where possible.

    SeeAlso:
        _alpha_blend_simple - alternative implementation

    Example:
        >>> rng = np.random.RandomState(0)
        >>> rgb1, rgb2 = rng.rand(10, 10, 3), rng.rand(10, 10, 3)
        >>> alpha1, alpha2 = rng.rand(10, 10), rng.rand(10, 10)
        >>> f1, f2 = _alpha_blend_inplace(rgb1, alpha1, rgb2, alpha2)
        >>> s1, s2 = _alpha_blend_simple(rgb1, alpha1, rgb2, alpha2)
        >>> assert np.all(f1 == s1) and np.all(f2 == s2)
        >>> alpha1, alpha2 = np.zeros((10, 10)), np.zeros((10, 10))
        >>> f1, f2 = _alpha_blend_inplace(rgb1, alpha1, rgb2, alpha2)
        >>> s1, s2 = _alpha_blend_simple(rgb1, alpha1, rgb2, alpha2)
        >>> assert np.all(f1 == s1) and np.all(f2 == s2)
    """
    rgb3 = np.empty_like(rgb1)
    temp_rgb = np.empty_like(rgb1)
    alpha3 = np.empty_like(alpha1)
    temp_alpha = np.empty_like(alpha1)

    # hold (1 - alpha1)
    np.subtract(1, alpha1, out=temp_alpha)

    # alpha3
    np.copyto(dst=alpha3, src=temp_alpha)
    np.multiply(alpha2, alpha3, out=alpha3)
    np.add(alpha1, alpha3, out=alpha3)

    # numer1
    np.multiply(rgb1, alpha1[..., None], out=rgb3)

    # numer2
    np.multiply(alpha2, temp_alpha, out=temp_alpha)
    np.multiply(rgb2, temp_alpha[..., None], out=temp_rgb)

    # (numer1 + numer2)
    np.add(rgb3, temp_rgb, out=rgb3)

    # removing errstate is actually a significant speedup
    with np.errstate(invalid='ignore'):
        np.divide(rgb3, alpha3[..., None], out=rgb3)
    if not np.all(alpha3):
        rgb3[alpha3 == 0] = 0
    return rgb3, alpha3


def _alpha_blend_numexpr1(rgb1, alpha1, rgb2, alpha2):
    """ Alternative. Not well optimized """
    import numexpr
    alpha1_ = alpha1[..., None]  # NOQA
    alpha2_ = alpha2[..., None]  # NOQA
    alpha3 = numexpr.evaluate('alpha1 + alpha2 * (1.0 - alpha1)')
    alpha3_ = alpha3[..., None]  # NOQA
    rgb3 = numexpr.evaluate('((rgb1 * alpha1_) + (rgb2 * alpha2_ * (1.0 - alpha1_))) / alpha3_')
    rgb3[alpha3 == 0] = 0


def _alpha_blend_numexpr2(rgb1, alpha1, rgb2, alpha2):
    """ Alternative. Not well optimized """
    import numexpr
    c_alpha1 = numexpr.evaluate('1.0 - alpha1')
    alpha3 = numexpr.evaluate('alpha1 + alpha2 * c_alpha1')

    c_alpha1_ = c_alpha1[..., None]  # NOQA
    alpha1_ = alpha1[..., None]  # NOQA
    alpha2_ = alpha2[..., None]  # NOQA
    alpha3_ = alpha3[..., None]  # NOQA

    numer1 = numexpr.evaluate('rgb1 * alpha1_')  # NOQA
    numer2 = numexpr.evaluate('rgb2 * (alpha2_ * c_alpha1_)')  # NOQA
    with np.errstate(invalid='ignore'):
        rgb3 = numexpr.evaluate('(numer1 + numer2) / alpha3_')
    rgb3[alpha3 == 0] = 0
    return rgb3, alpha3


def ensure_alpha_channel(img, alpha=1.0, dtype=np.float32, copy=False):
    """
    Returns the input image with 4 channels.

    Args:
        img (ndarray):
            an image with shape [H, W], [H, W, 1], [H, W, 3], or [H, W, 4].

        alpha (float | ndarray):
            default scalar value for missing alpha channel, or
            an ndarray with the same height / width to use explicitly.

        dtype (type):
            The final output dtype. Should be numpy.float32 or numpy.float64.

        copy (bool):
            always copy if True, else copy if needed.

    Returns:
        ndarray: an image with specified dtype with shape [H, W, 4].

    Raises:
        ValueError - if the input image does not have 1, 3, or 4 input channels
            or if the image cannot be converted into a float01 representation

    Example:
        >>> # Demo with a scalar default alpha value
        >>> import kwimage
        >>> data0 = np.zeros((5, 5))
        >>> data1 = np.zeros((5, 5, 1))
        >>> data2 = np.zeros((5, 5, 3))
        >>> data3 = np.zeros((5, 5, 4))
        >>> ensured0 = kwimage.ensure_alpha_channel(data0, alpha=0.5)
        >>> ensured1 = kwimage.ensure_alpha_channel(data1, alpha=0.5)
        >>> ensured2 = kwimage.ensure_alpha_channel(data2, alpha=0.5)
        >>> ensured3 = kwimage.ensure_alpha_channel(data3, alpha=0.5)
        >>> assert np.all(ensured0[..., 3] == 0.5), 'should have been populated'
        >>> assert np.all(ensured1[..., 3] == 0.5), 'should have been populated'
        >>> assert np.all(ensured2[..., 3] == 0.5), 'should have been populated'
        >>> assert np.all(ensured3[..., 3] == 0.0), 'last image already had alpha'

    Example:
        >>> import kwimage
        >>> # Demo with a explicit alpha channel
        >>> alpha = np.random.rand(5, 5)
        >>> data0 = np.zeros((5, 5))
        >>> data1 = np.zeros((5, 5, 1))
        >>> data2 = np.zeros((5, 5, 3))
        >>> data3 = np.zeros((5, 5, 4))
        >>> ensured0 = kwimage.ensure_alpha_channel(data0, alpha=alpha)
        >>> ensured1 = kwimage.ensure_alpha_channel(data1, alpha=alpha)
        >>> ensured2 = kwimage.ensure_alpha_channel(data2, alpha=alpha)
        >>> ensured3 = kwimage.ensure_alpha_channel(data3, alpha=alpha)
        >>> assert np.all(ensured0[..., 3] == alpha), 'should have been populated'
        >>> assert np.all(ensured1[..., 3] == alpha), 'should have been populated'
        >>> assert np.all(ensured2[..., 3] == alpha), 'should have been populated'
        >>> assert np.all(ensured3[..., 3] == 0.0), 'last image already had alpha'
    """
    img = im_core.ensure_float01(img, dtype=dtype, copy=copy)
    c = im_core.num_channels(img)
    if c == 4:
        return img
    else:
        if isinstance(alpha, np.ndarray):
            alpha_channel = alpha
        else:
            alpha_channel = np.full(img.shape[0:2], fill_value=alpha, dtype=img.dtype)
        if c == 3:
            return np.dstack([img, alpha_channel])
        elif c == 1:
            return np.dstack([img, img, img, alpha_channel])
        else:
            raise ValueError(
                'Cannot ensure alpha. Input image has c={} channels'.format(c))



import itertools as it
import numpy as np
import ubelt as ub


def test_from_mask_rle_translate():
    """
    Ensure that RLE translation works on a variety of data
    """
    from kwimage.structs.mask import Mask
    from kwimage.structs.mask import MaskFormat  # NOQA

    SMALL = True

    if SMALL:
        # Only test a restricted number of examples to make things faster for
        # automated tests.
        N, M = 2, 3
        offset_choices = [
            (1, 1), (0, 0), (-1, -1),
        ]
        shape_choices = [
            (4, 4), (N, M), (2, 2),
        ]
    else:
        N, M = 3, 5
        offset_choices = [
            (1, 2), (0, 0), (-1, 0), (0, -1), (-1, -2),
        ]
        shape_choices = [
            (N + 2, M + 2),
            (N, M),
            (N - 2, M - 2),
            (N - 1, M),
            (N, M - 1),
        ]

    # Ensure this works all possible NxM binary images
    pixel_choices = [[0, 1]] * (N * M)
    total = np.prod(list(map(len, pixel_choices)))

    iter_ = it.product(*pixel_choices)
    iter_ = ub.ProgIter(iter_, total=total, desc='testing RLE translate')
    for choice in iter_:
        mask = np.array(choice, dtype=np.uint8).reshape(N, M)
        for offset in offset_choices:
            for shape in shape_choices:
                self1 = Mask.from_mask(mask, offset, shape, 'naive')
                self2 = Mask.from_mask(mask, offset, shape, 'faster')

                m1 = self1.to_c_mask().data
                m2 = self2.to_c_mask().data

                assert np.all(m1 == m2)

                # if False:
                #     import kwimage
                #     encoded = kwimage.encode_run_length(mask, binary=True, order='F')
                #     encoded['size'] = encoded['shape']
                #     self = Mask(encoded, MaskFormat.ARRAY_RLE)


def test_rle_translate():
    """
    Ensure the functional RLE encoders / transformers work on C and F data
    """
    import kwimage

    SMALL = True

    if SMALL:
        # Only test a restricted number of examples to make things faster for
        # automated tests.
        N, M = 2, 3
        offset_choices = [
            (1, 1), (0, 0), (-1, -1),
        ]
        shape_choices = [
            (4, 4), (N, M), (2, 2),
        ]
    else:
        N, M = 3, 5
        offset_choices = [
            (1, 2), (0, 0), (-1, 0), (0, -1), (-1, -2),
        ]
        shape_choices = [
            (N + 2, M + 2),
            (N, M),
            (N - 2, M - 2),
            (N - 1, M),
            (N, M - 1),
        ]

    # Ensure this works all possible NxM binary images
    pixel_choices = [[0, 1]] * (N * M)
    total = np.prod(list(map(len, pixel_choices)))

    iter_ = it.product(*pixel_choices)
    iter_ = ub.ProgIter(iter_, total=total, desc='testing RLE translate')
    for choice in iter_:
        mask = np.array(choice, dtype=np.uint8).reshape(N, M)
        c_rle = kwimage.encode_run_length(mask, order='C', binary=True)
        f_rle = kwimage.encode_run_length(mask, order='F', binary=True)
        for offset in offset_choices:
            for shape in shape_choices:
                c_rle2 = kwimage.rle_translate(c_rle, offset, shape)
                f_rle2 = kwimage.rle_translate(f_rle, offset, shape)

                m1 = kwimage.decode_run_length(**c_rle2)
                m2 = kwimage.decode_run_length(**f_rle2)
                assert np.all(m1 == m2)



from typing import Tuple
from torch import Tensor
from numpy.typing import ArrayLike
from typing import Sequence
from _typeshed import Incomplete


def warp_tensor(inputs: Tensor,
                mat: Tensor,
                output_dims: Tuple[int, ...],
                mode: str = 'bilinear',
                padding_mode: str = 'zeros',
                isinv: bool = False,
                ishomog: bool | None = None,
                align_corners: bool = False,
                new_mode: bool = ...) -> Tensor:
    ...


def subpixel_align(dst, src, index, interp_axes: Incomplete | None = ...):
    ...


def subpixel_set(dst: ArrayLike,
                 src: ArrayLike,
                 index: Tuple[slice],
                 interp_axes: tuple | None = None):
    ...


def subpixel_accum(dst: ArrayLike,
                   src: ArrayLike,
                   index: Tuple[slice],
                   interp_axes: tuple | None = None):
    ...


def subpixel_maximum(dst: ArrayLike,
                     src: ArrayLike,
                     index: Tuple[slice],
                     interp_axes: tuple | None = None):
    ...


def subpixel_minimum(dst: ArrayLike,
                     src: ArrayLike,
                     index: Tuple[slice],
                     interp_axes: tuple | None = None):
    ...


def subpixel_slice(inputs: ArrayLike, index: Tuple[slice]):
    ...


def subpixel_translate(inputs: ArrayLike,
                       shift: Sequence,
                       interp_axes: Sequence | None = None,
                       output_shape: tuple | None = None):
    ...


def warp_points(matrix: ArrayLike, pts: ArrayLike, homog_mode: str = 'divide'):
    ...


def remove_homog(pts, mode: str = ...):
    ...


def add_homog(pts):
    ...


def subpixel_getvalue(img: ArrayLike,
                      pts: ArrayLike,
                      coord_axes: Sequence | None = None,
                      interp: str = 'bilinear',
                      bordermode: str = 'edge'):
    ...


def subpixel_setvalue(img: ArrayLike,
                      pts: ArrayLike,
                      value: ArrayLike,
                      coord_axes: Sequence | None = None,
                      interp: str = 'bilinear',
                      bordermode: str = 'edge'):
    ...




def test_warp_affine():
    from kwimage._backend_info import _have_cv2
    if not _have_cv2():
        import pytest
        pytest.skip('requires cv2')
    # Demo how of how we also handle masked arrays
    import kwimage
    import numpy as np
    _image = kwimage.grab_test_image('pm5644')
    _image = kwimage.ensure_float01(_image)
    _image[100:200, 400:700] = np.nan
    mask = np.isnan(_image)
    data = np.nan_to_num(_image)
    image = np.ma.MaskedArray(data=data, mask=mask)
    transform = kwimage.Affine.coerce(scale=0.05, offset=10.5, theta=0.3, shearx=0.2)
    warped1 = kwimage.warp_affine(image, transform, dsize='positive', antialias=1, interpolation='linear')
    assert isinstance(warped1, np.ma.MaskedArray)
    warped2 = kwimage.warp_affine(image, transform, dsize='positive', antialias=0)
    print('warped1.shape = {!r}'.format(warped1.shape))
    print('warped2.shape = {!r}'.format(warped2.shape))
    assert warped2.shape == warped1.shape

    if 0:
        # xdoctest: +REQUIRES(--show)
        import kwplot
        kwplot.autompl()
        pnum_ = kwplot.PlotNums(nRows=1, nCols=2)

        canvas1 = kwimage.nodata_checkerboard(warped1, square_shape=1)
        canvas2 = kwimage.nodata_checkerboard(warped2, square_shape=1)
        kwplot.imshow(canvas1, pnum=pnum_(), title='antialias=True')
        kwplot.imshow(canvas2, pnum=pnum_(), title='antialias=False')
        kwplot.show_if_requested()


def test_warp_affine_with_nan_border():
    from kwimage._backend_info import _have_cv2
    if not _have_cv2():
        import pytest
        pytest.skip('requires cv2')
    import kwimage
    import numpy as np
    img = kwimage.ensure_float01(kwimage.grab_test_image())
    M = kwimage.Affine.affine(theta=np.pi / 8)

    # TODO: ported to kwarray, use that later
    def equal_with_nan(a1, a2):
        """
        Numpy has array_equal with ``equal_nan=True``, but this is elementwise

        Args:
            a1 (ArrayLike): input array
            a2 (ArrayLike): input array
        """
        a1, a2 = np.asarray(a1), np.asarray(a2)
        a1nan, a2nan = np.isnan(a1), np.isnan(a2)
        nan_sameness = a1nan == a2nan
        value_sameness = (a1 == a2)
        # If they are actually the same, they should be value same xor nansame.
        flags = value_sameness ^ nan_sameness
        return flags

    # Explicit tuple should be for each channel
    border_value = (np.nan, 0, np.nan)
    warped = kwimage.warp_affine(img, M, border_value=border_value)
    nanned_pixels = warped[np.isnan(warped).any(axis=2)]
    assert nanned_pixels.size
    assert not warped[np.isnan(warped).any(axis=2)].all()
    assert equal_with_nan(nanned_pixels, border_value).all(axis=1).any()

    border_value = (np.nan, 0, 0)
    warped = kwimage.warp_affine(img, M, border_value=border_value)
    nanned_pixels = warped[np.isnan(warped).any(axis=2)]
    assert nanned_pixels.size
    assert not warped[np.isnan(warped).any(axis=2)].all()
    assert equal_with_nan(nanned_pixels, border_value).all(axis=1).any()

    border_value = np.nan
    warped = kwimage.warp_affine(img, M, border_value=border_value)
    nanned_pixels = warped[np.isnan(warped).any(axis=2)]
    assert nanned_pixels.size
    assert warped[np.isnan(warped).any(axis=2)].all()

    # Case with a scalar and a 2D image.
    border_value = np.nan
    img0 = img[:, :, 0]
    warped = kwimage.warp_affine(img0, M, border_value=border_value)
    nanned_pixels = warped[np.isnan(warped)]
    assert nanned_pixels.size


def test_warp_affine_with_many_chans():
    from kwimage._backend_info import _have_cv2
    if not _have_cv2():
        import pytest
        pytest.skip('requires cv2')
    import kwimage
    import numpy as np
    img = np.random.rand(5, 5, 4)
    M = kwimage.Affine.affine(theta=np.pi / 8)

    img = np.random.rand(8, 8, 5)
    M = kwimage.Affine.affine(theta=np.pi / 8)
    warped = kwimage.warp_affine(img, M, border_value=np.nan)
    warped



"""
Generic segmentation object that can use either a Mask or (Multi)Polygon
backend.
"""
# from kwimage.structs import _generic
import numpy as np
import numbers
from . import _generic
import ubelt as ub


class _WrapperObject(ub.NiceRepr):

    def __nice__(self):
        return self.data.__nice__()

    def draw(self, *args, **kw):
        return self.data.draw(*args, **kw)

    def draw_on(self, *args, **kw):
        """
        See help(self.data.draw_on)
        """
        return self.data.draw_on(*args, **kw)

    def warp(self, *args, **kw):
        return self.data.warp(*args, **kw)

    def translate(self, *args, **kw):
        return self.data.translate(*args, **kw)

    def scale(self, *args, **kw):
        return self.data.scale(*args, **kw)

    def to_coco(self, *args, **kw):
        return self.data.to_coco(*args, **kw)

    def numpy(self, *args, **kw):
        return self.data.numpy(*args, **kw)

    def tensor(self, *args, **kw):
        return self.data.tensor(*args, **kw)


class Segmentation(_WrapperObject):
    """
    Either holds a MultiPolygon, Polygon, or Mask

    Args:
        data (object): the underlying object
        format (str): either 'mask', 'polygon', or 'multipolygon'
    """
    def __init__(self, data, format=None):
        self.data = data
        self.format = format

    @classmethod
    def random(cls, rng=None):
        """
        Example:
            >>> # xdoctest: +REQUIRES(module:cv2)
            >>> self = Segmentation.random()
            >>> print('self = {!r}'.format(self))
            >>> # xdoctest: +REQUIRES(--show)
            >>> import kwplot
            >>> kwplot.autompl()
            >>> kwplot.figure(fnum=1, doclf=True)
            >>> self.draw()
            >>> kwplot.show_if_requested()
        """

        import kwarray
        import kwimage
        rng = kwarray.ensure_rng(rng)
        if rng.rand() > 0.5:
            data = kwimage.Polygon.random()
        else:
            data = kwimage.Mask.random()
        return cls.coerce(data)

    def to_multi_polygon(self):
        return self.data.to_multi_polygon()

    def to_mask(self, dims=None, pixels_are='points'):
        return self.data.to_mask(dims=dims, pixels_are=pixels_are)

    def box(self):
        return self.data.box()

    @property
    def area(self):
        return self.data.area

    @property
    def meta(self):
        return self.data.meta

    @classmethod
    def coerce(cls, data, dims=None):
        import kwimage
        if _generic._isinstance2(data, kwimage.Segmentation):
            self = data
        elif _generic._isinstance2(data, kwimage.Mask):
            self = Segmentation(data, 'mask')
        elif _generic._isinstance2(data, kwimage.Polygon):
            self = Segmentation(data, 'polygon')
        elif _generic._isinstance2(data, kwimage.MultiPolygon):
            self = Segmentation(data, 'multipolygon')
        else:
            data = _coerce_coco_segmentation(data, dims=dims)
            if data is None:
                return None
            self = cls.coerce(data, dims=dims)
        return self


class SegmentationList(_generic.ObjectList):
    """
    Store and manipulate multiple segmentations (masks or polygons), usually
    within the same image
    """

    def to_polygon_list(self):
        """
        Converts all mask objects to multi-polygon objects
        """
        import kwimage
        new = kwimage.PolygonList([
            None if item is None else item.to_multi_polygon()
            for item in self
        ])
        return new

    def to_mask_list(self, dims=None, pixels_are='points'):
        """
        Converts all mask objects to multi-polygon objects
        """
        import kwimage
        new = kwimage.MaskList([
            None if item is None else item.to_mask(dims=dims, pixels_are=pixels_are)
            for item in self
        ])
        return new

    def to_segmentation_list(self):
        return self

    @classmethod
    def coerce(cls, data, none_policy='raise'):
        """
        Interpret data as a list of Segmentations

        Args:
            none_policy (str):
                Determines how to handle None inputs.
                Can be: 'return-None', or 'raise'.
        """
        if isinstance(data, (list, _generic.ObjectList)):
            data = [None if item is None else Segmentation.coerce(item)
                    for item in data]
        else:
            if data is None:
                return _handle_null_policy(none_policy)
            else:
                raise TypeError(data)
        self = cls(data)
        return self


def _handle_null_policy(policy, ex_type=TypeError,
                        ex_msg='cannot accept null input'):
    """
    For handling a nan or None policy.

    Args:
        policy (str):
            How null inputs are handled. Can be:
                'return-None': returns None
                'return-nan': returns nan
                'raise': raises an error

        ex_type (type): Exception type to raise if policy is raise

        ex_msg (msg): Exception arguments

    TODO: rectify with similar logic in kwutil/util_time
    """
    if policy == 'return-None':
        return None
    elif policy == 'return-nan':
        return float('nan')
    elif policy == 'raise':
        raise ex_type(ex_msg)
    else:
        raise KeyError(ub.paragraph(
            f'''
            Unknown null policy={policy!r}.
            Valid choices are "return-None", "return-nan", and "raise".
            '''))


def _coerce_coco_segmentation(data, dims=None):
    """
    Attempts to auto-inspect the format of segmentation data

    Args:
        data : the data to coerce

             2D-C-ndarray -> C_MASK
             2D-F-ndarray -> F_MASK

             Dict(counts=bytes) -> BYTES_RLE
             Dict(counts=ndarray) -> ARRAY_RLE

             Dict(exterior=ndarray) -> ARRAY_RLE

             # List[List[int]] -> Polygon
             List[int] -> Polygon
             List[Dict] -> MultPolygon

        dims (Tuple): required for certain formats like polygons
            height / width of the source image

    TODO:
        - [ ] Handle WKT

    Returns:
        Mask | Polygon | MultiPolygon | Segmentation - depending on which is appropriate

    Example:
        >>> segmentation = {'size': [5, 9], 'counts': ';?1B10O30O4'}
        >>> dims = (9, 5)
        >>> raw_mask = (np.random.rand(32, 32) > .5).astype(np.uint8)
        >>> _coerce_coco_segmentation(segmentation)
        >>> _coerce_coco_segmentation(raw_mask)

        >>> coco_polygon = [
        >>>     np.array([[3, 0],[2, 1],[2, 4],[4, 4],[4, 3],[7, 0]]),
        >>>     np.array([[2, 1],[2, 2],[4, 2],[4, 1]]),
        >>> ]
        >>> self = _coerce_coco_segmentation(coco_polygon, dims)
        >>> print('self = {!r}'.format(self))
        >>> coco_polygon = [
        >>>     np.array([[3, 0],[2, 1],[2, 4],[4, 4],[4, 3],[7, 0]]),
        >>> ]
        >>> self = _coerce_coco_segmentation(coco_polygon, dims)
        >>> print('self = {!r}'.format(self))
    """
    import kwimage
    from kwimage.structs.mask import MaskFormat
    if isinstance(data, np.ndarray):
        # INPUT TYPE: RAW MASK
        if dims is not None:
            assert dims == data.shape[0:2]
        if data.flags['F_CONTIGUOUS']:
            self = kwimage.Mask(data, MaskFormat.F_MASK)
        else:
            self = kwimage.Mask(data, MaskFormat.C_MASK)
    elif isinstance(data, dict):
        if 'counts' in data:
            # INPUT TYPE: COCO RLE DICTIONARY
            if dims is not None:
                data_shape = data.get('dims', data.get('shape', data.get('size', None)))
                if data_shape is None:
                    data['shape'] = data_shape
                else:
                    assert tuple(map(int, dims)) == tuple(map(int, data_shape)), (
                        '{} {}'.format(dims, data_shape))
            if isinstance(data['counts'], (str, bytes)):
                self = kwimage.Mask(data, MaskFormat.BYTES_RLE)
            else:
                self = kwimage.Mask(data, MaskFormat.ARRAY_RLE)
        elif 'exterior' in data:
            # TODO: kwimage.Polygon.from_coco
            self = kwimage.Polygon(**data)
            # raise NotImplementedError('explicit polygon coerce')
        elif 'type' in data:
            if data['type'] == 'Polygon':
                self = kwimage.Polygon.from_geojson(data)
            elif data['type'] == 'MultiPolygon':
                self = kwimage.MultiPolygon.from_geojson(data)
            else:
                raise NotImplementedError(data['type'])
        else:
            raise TypeError('Unable to interpret dictionary format {}'.format(data))
    elif isinstance(data, list):
        # THIS IS NOT AN IDEAL FORMAT. IDEALLY WE WILL MODIFY COCO TO USE
        # DICTIONARIES FOR POLYGONS, WHICH ARE UNAMBIGUOUS
        if len(data) == 0:
            self = None
        else:
            first = ub.peek(data)
            if isinstance(first, dict):
                # TODO: kwimage.MultiPolygon.from_coco
                self = kwimage.MultiPolygon(
                    [kwimage.Polygon(**item) for item in data])
            elif isinstance(first, numbers.Number):
                # TODO: kwimage.Polygon.from_coco
                exterior = np.array(data).reshape(-1, 2)
                self = kwimage.Polygon(exterior=exterior)
            elif isinstance(first, list):
                # TODO: kwimage.MultiPolygon.from_coco
                poly_list = [kwimage.Polygon(exterior=np.array(item).reshape(-1, 2))
                             for item in data]
                if len(poly_list) == 1:
                    self = poly_list[0]
                else:
                    self = kwimage.MultiPolygon(poly_list)
            elif isinstance(first, np.ndarray):
                poly_list = [kwimage.Polygon(exterior=item.reshape(-1, 2))
                             for item in data]
                if len(poly_list) == 1:
                    self = poly_list[0]
                else:
                    self = kwimage.MultiPolygon(poly_list)
            else:
                raise TypeError('Unable to interpret list format {}'.format(data))
    elif isinstance(data, (kwimage.Polygon, kwimage.MultiPolygon, kwimage.Mask, kwimage.Segmentation)):
        self = data
    else:
        from shapely.geometry.polygon import Polygon
        from shapely.geometry.multipolygon import MultiPolygon
        if isinstance(data, MultiPolygon):
            self = kwimage.MultiPolygon.from_shapely(data)
        elif isinstance(data, Polygon):
            self = kwimage.Polygon.from_shapely(data)
        else:
            raise TypeError(f'Unable to coerce type={type(data)!r} into a segmentation. data={data!r}')
    return self



"""
Generic Non-Maximum Suppression API with efficient backend implementations
"""
import sys
import numpy as np
import ubelt as ub
import warnings
import kwarray


def daq_spatial_nms(ltrb, scores, diameter, thresh, max_depth=6,
                    stop_size=2048, recsize=2048, impl='auto', device_id=None):
    """
    Divide and conquer speedup non-max-supression algorithm for when bboxes
    have a known max size

    Args:
        ltrb (ndarray): boxes in (tlx, tly, brx, bry) format

        scores (ndarray): scores of each box

        diameter (int | Tuple[int, int]): Distance from split point to
            consider rectification. If specified as an integer, then number
            is used for both height and width. If specified as a tuple, then
            dims are assumed to be in [height, width] format.

        thresh (float): iou threshold. Boxes are removed if they overlap
            greater than this threshold. 0 is the most strict, resulting in the
            fewest boxes, and 1 is the most permissive resulting in the most.

        max_depth (int): maximum number of times we can divide and conquer

        stop_size (int): number of boxes that triggers full NMS computation

        recsize (int): number of boxes that triggers full NMS recombination

        impl (str): algorithm to use. Defaults to "auto". Can be "python",
            "cython_cpu", "gpu", "torch", or "torchvision".

    Note:

        # TODO: Look Into

        # Didn't read yet but it seems similar

        http://www.cyberneum.de/fileadmin/user_upload/files/publications/CVPR2010-Lampert_[0].pdf

        https://www.researchgate.net/publication/220929789_Efficient_Non-Maximum_Suppression

        # This seems very similar

        https://projet.liris.cnrs.fr/m2disco/pub/Congres/2006-ICPR/DATA/C03_0406.PDF

    Example:
        >>> import kwimage
        >>> # Make a bunch of boxes with the same width and height
        >>> #boxes = kwimage.Boxes.random(230397, scale=1000, format='cxywh')
        >>> boxes = kwimage.Boxes.random(237, scale=1000, format='cxywh')
        >>> boxes.data.T[2] = 10
        >>> boxes.data.T[3] = 10
        >>> #
        >>> ltrb = boxes.to_ltrb().data.astype(np.float32)
        >>> scores = np.arange(0, len(ltrb)).astype(np.float32)
        >>> #
        >>> n_megabytes = (ltrb.size * ltrb.dtype.itemsize) / (2 ** 20)
        >>> print('n_megabytes = {!r}'.format(n_megabytes))
        >>> #
        >>> thresh = iou_thresh = 0.01
        >>> impl = 'auto'
        >>> max_depth = 20
        >>> diameter = 10
        >>> stop_size = 2000
        >>> recsize = 500
        >>> #
        >>> import ubelt as ub
        >>> #
        >>> with ub.Timer(label='daq'):
        >>>     keep1 = daq_spatial_nms(ltrb, scores,
        >>>         diameter=diameter, thresh=thresh, max_depth=max_depth,
        >>>         stop_size=stop_size, recsize=recsize, impl=impl)
        >>> #
        >>> with ub.Timer(label='full'):
        >>>     keep2 = non_max_supression(ltrb, scores,
        >>>         thresh=thresh, impl=impl)
        >>> #
        >>> # Due to the greedy nature of the algorithm, there will be slight
        >>> # differences in results, but they will be mostly similar.
        >>> similarity = len(set(keep1) & set(keep2)) / len(set(keep1) | set(keep2))
        >>> print('similarity = {!r}'.format(similarity))
    """
    def _rectify(ltrb, both_keep, needs_rectify):
        if len(needs_rectify) == 0:
            keep = sorted(both_keep)
        else:
            nr_arr = np.array(sorted(needs_rectify))
            nr = needs_rectify
            bk = set(both_keep)
            rectified_keep = non_max_supression(
                ltrb[nr_arr], scores[nr_arr], thresh=thresh,
                impl=impl, device_id=device_id)
            rk = set(nr_arr[rectified_keep])
            keep = sorted((bk - nr) | rk)
        return keep

    def _recurse(ltrb, scores, dim, depth, diameter_wh):
        """
        Args:
            dim (int): flips between 0 and 1
            depth (int): recursion depth
        """
        # print('recurse')
        n_boxes = len(ltrb)
        if depth >= max_depth or n_boxes < stop_size:
            # print('n_boxes = {!r}'.format(n_boxes))
            # print('depth = {!r}'.format(depth))
            # print('stop')
            keep = non_max_supression(ltrb, scores, thresh=thresh, impl=impl)
            both_keep = sorted(keep)
            needs_rectify = set()
        else:
            # Break up the NMS into two subproblems.
            middle = np.median(ltrb.T[dim])
            left_flags = ltrb.T[dim] < middle
            right_flags = ~left_flags

            left_idxs = np.where(left_flags)[0]
            right_idxs = np.where(right_flags)[0]

            left_scores = scores[left_idxs]
            left_ltrb = ltrb[left_idxs]

            right_scores = scores[right_idxs]
            right_ltrb = ltrb[right_idxs]

            next_depth = depth + 1
            next_dim = 1 - dim

            # Solve each subproblem
            left_keep_, lrec_ = _recurse(
                left_ltrb, left_scores, depth=next_depth, dim=next_dim,
                diameter_wh=diameter_wh)

            right_keep_, rrec_ = _recurse(
                right_ltrb, right_scores, depth=next_depth, dim=next_dim,
                diameter_wh=diameter_wh)

            # Recombine the results (note that because we have a diameter_wh,
            # we have to check less results)
            rrec = set(right_idxs[sorted(rrec_)])
            lrec = set(left_idxs[sorted(lrec_)])

            left_keep = left_idxs[left_keep_]
            right_keep = right_idxs[right_keep_]

            both_keep = np.hstack([left_keep, right_keep])
            both_keep.sort()

            dist_to_middle = np.abs(ltrb[both_keep].T[dim] - middle)

            # Find all surviving boxes that are close to the midpoint.  We will
            # need to recheck these because they may overlap, but they also may
            # have been split into different subproblems.
            rectify_flags = dist_to_middle < diameter_wh[dim]

            needs_rectify = set(both_keep[rectify_flags])
            needs_rectify.update(rrec)
            needs_rectify.update(lrec)

            nrec = len(needs_rectify)
            # print('nrec = {!r}'.format(nrec))
            if nrec > recsize:
                both_keep = _rectify(ltrb, both_keep, needs_rectify)
                needs_rectify = set()
        return both_keep, needs_rectify

    if not ub.iterable(diameter):
        diameter_wh = [diameter, diameter]
    else:
        diameter_wh = diameter[::-1]

    depth = 0
    dim = 0
    both_keep, needs_rectify = _recurse(ltrb, scores, dim=dim, depth=depth,
                                        diameter_wh=diameter_wh)
    keep = _rectify(ltrb, both_keep, needs_rectify)
    return keep


_impls = None


class _NMS_Impls():
    # TODO: could make this prettier
    def __init__(self):
        self._funcs = None

    def _lazy_init(self):
        _funcs = {}
        from kwimage import _internal
        try:
            import torch
        except ImportError:
            torch = None

        # These are pure python and should always be available
        from kwimage.algo._nms_backend import py_nms
        from kwimage.algo._nms_backend import torch_nms
        _funcs['numpy'] = py_nms.py_nms

        try:
            from packaging.version import parse as LooseVersion
        except ImportError:
            from distutils.version import LooseVersion
        recent_numpy = LooseVersion(np.__version__) >= LooseVersion('1.20.0')

        if torch is not None and recent_numpy:
            _funcs['torch'] = torch_nms.torch_nms

            if not _internal.KWIMAGE_DISABLE_TORCHVISION_NMS:
                # The torchvision _C libraray may cause segfaults, which is
                # why we have an option to disable even trying it
                try:
                    # TODO: torchvision impl might be the best, need to test
                    # from torchvision import _C as C  # NOQA
                    import torchvision
                    _funcs['torchvision'] = torchvision.ops.nms
                except (ImportError, UnicodeDecodeError) as ex:
                    warnings.warn(
                        'optional torchvision C nms is not available: {}'.format(
                            str(ex)))

        if recent_numpy:
            # Only use cython extensions if numpy is > 1.20.
            # Otherwise there seems to be a
            # "Fatal Python error: Illegal instruction" that can occur
            # I dont know why this is yet
            # See https://gitlab.kitware.com/computer-vision/kwimage/-/jobs/6061311
            # for an example of when this last occurred
            try:
                if not _internal.KWIMAGE_DISABLE_C_EXTENSIONS:
                    from kwimage_ext.algo._nms_backend import cpu_nms
                    _funcs['cython_cpu'] = cpu_nms.cpu_nms
            except Exception as ex:
                warnings.warn(
                    'optional cpu_nms is not available: {}'.format(str(ex)))
            try:
                if not _internal.KWIMAGE_DISABLE_C_EXTENSIONS:
                    if torch is not None and torch.cuda.is_available():
                        from kwimage_ext.algo._nms_backend import gpu_nms
                        _funcs['cython_gpu'] = gpu_nms.gpu_nms
                        # NOTE: GPU is not the fastests on all systems.
                        # See the benchmarks for more info.
                        # ~/code/kwimage/dev/bench_nms.py
            except Exception as ex:
                warnings.warn(
                    'optional gpu_nms is not available: {}'.format(str(ex)))
        self._funcs = _funcs
        self._valid = frozenset(_impls._funcs.keys())


_impls = _NMS_Impls()


def available_nms_impls():
    """
    List available values for the `impl` kwarg of `non_max_supression`

    CommandLine:
        xdoctest -m kwimage.algo.algo_nms available_nms_impls

    Example:
        >>> from kwimage.algo.algo_nms import *  # NOQA
        >>> impls = available_nms_impls()
        >>> assert 'numpy' in impls
        >>> print('impls = {!r}'.format(impls))
    """
    if not _impls._funcs:
        _impls._lazy_init()
    return list(_impls._funcs.keys())


# @ub.memoize
def _heuristic_auto_nms_impl(code, num, valid=None):
    """
    Defined with help from ``~/code/kwimage/dev/bench/bench_nms.py``

    Args:
        code (str): text that indicates which type of data you have
            tensor0 is a tensor on a cuda device, tensor is on the cpu, and
            numpy is a ndarray.

        num (int): number of boxes you have to supress.

        valid (List[str]): the list of valid implementations, an error will be
            raised if heuristic preferences do not intersect with this list.

    Ignore:
        _impls._funcs
        valid_pref = ub.oset(preference) & set(_impls._funcs.keys())
        python ~/code/kwimage/dev/bench/bench_nms.py --show --small-boxes --thresh=0.6
    """
    if code not in {'tensor0', 'tensor', 'ndarray'}:
        raise KeyError(code)

    if num <= 10:
        if code == 'tensor0':
            # dict(cython_cpu=4118.4, torchvision=3042.5, cython_gpu=2244.4, torch=841.9)
            preference = ['cython_cpu', 'torchvision', 'cython_gpu', 'torch']
        if code == 'tensor':
            # dict(torchvision=5857.1, cython_gpu=3058.1)
            preference = ['torchvision', 'cython_gpu', 'torch', 'numpy']
        if code == 'ndarray':
            # dict(cython_cpu=12226.1, numpy=7759.1, cython_gpu=3679.0, torch=1786.2)
            preference = ['cython_cpu', 'numpy', 'cython_gpu', 'torch']
    elif num <= 100:
        if code == 'tensor0':
            # dict(cython_cpu=4160.7, torchvision=3089.9, cython_gpu=2261.8, torch=846.8)
            preference = ['cython_cpu', 'torchvision', 'cython_gpu', 'torch', 'numpy']
        if code == 'tensor':
            # dict(torchvision=5875.3, cython_gpu=3076.9)
            preference = ['torchvision', 'cython_gpu', 'torch', 'numpy']
        if code == 'ndarray':
            # dict(cython_cpu=12256.7, cython_gpu=3702.9, numpy=2311.3, torch=1738.0)
            preference = ['cython_cpu', 'cython_gpu', 'numpy', 'torch']
    elif num <= 200:
        if code == 'tensor0':
            # dict(cython_cpu=3460.8, torchvision=2912.9, cython_gpu=2125.2, torch=782.4)
            preference = ['cython_cpu', 'torchvision', 'cython_gpu', 'torch']
        if code == 'tensor':
            # dict(torchvision=3394.6, cython_gpu=2641.2)
            preference = ['torchvision', 'cython_gpu', 'torch', 'numpy']
        if code == 'ndarray':
            # dict(cython_cpu=8220.6, cython_gpu=3114.5, torch=1240.7, numpy=309.5)
            preference = ['cython_cpu', 'cython_gpu', 'torch', 'numpy']
    elif num <= 300:
        if code == 'tensor0':
            # dict(torchvision=2647.1, cython_cpu=2264.9, cython_gpu=1915.5, torch=672.0)
            preference = ['torchvision', 'cython_cpu', 'cython_gpu', 'torch']
        if code == 'tensor':
            # dict(cython_gpu=2496.9, torchvision=1781.1)
            preference = ['cython_gpu', 'torchvision', 'torch', 'numpy']
        if code == 'ndarray':
            # dict(cython_cpu=4085.6, cython_gpu=2944.4, torch=799.8, numpy=173.0)
            preference = ['cython_cpu', 'cython_gpu', 'torch', 'numpy']
    else:
        if code == 'tensor0':
            # dict(torchvision=2585.5, cython_gpu=1868.7, cython_cpu=1650.6, torch=623.1)
            preference = ['torchvision', 'cython_gpu', 'cython_cpu', 'torch']
        if code == 'tensor':
            # dict(cython_gpu=2463.1, torchvision=1126.2)
            preference = ['cython_gpu', 'torchvision', 'torch', 'numpy']
        if code == 'ndarray':
            # dict(cython_gpu=2880.2, cython_cpu=2432.5, torch=511.9, numpy=114.0)
            preference = ['cython_gpu', 'cython_cpu', 'torch', 'numpy']

    if valid:
        valid_pref = ub.oset(preference) & valid
    else:
        valid_pref = preference

    if not valid_pref:
        raise Exception(
            'no valid nms algo: code={}, num={}, valid={}, preference={}, valid_pref={}'.format(
                code, num, valid, preference, valid_pref))

    impl = valid_pref[0]
    return impl


def non_max_supression(ltrb, scores, thresh, bias=0.0, classes=None,
                       impl='auto', device_id=None):
    """
    Non-Maximum Suppression - remove redundant bounding boxes

    Based on information from [CythonNMS]_ and [NMSPython]_.

    Args:
        ltrb (ndarray[Any, Float32]):
            Float32 array of shape Nx4 representing boxes in ltrb format

        scores (ndarray[Any, Float32]):
            Float32 array of shape N representing scores for each box

        thresh (float): iou threshold.
            Boxes are removed if they overlap greater than this threshold
            (i.e. Boxes are removed if iou > threshold).
            Thresh = 0 is the most strict, resulting in the fewest boxes, and 1
            is the most permissive resulting in the most.

        bias (float): bias for iou computation either 0 or 1

        classes (ndarray[Shape['*'], Int64] | None): integer classes.
            If specified NMS is done on a perclass basis.

        impl (str): implementation can be "auto", "python", "cython_cpu",
            "gpu", "torch", or "torchvision". Not all backends may be
            available, see :func:`kwimage.algo.algo_nms.available_nms_impls`
            for what is supported on your system.

        device_id (int): used if impl is gpu, device id to work on. If not
            specified `torch.cuda.current_device()` is used.

    Note:
        Using impl='cython_gpu' may result in an CUDA memory error that is not exposed
        to the python processes. In other words your program will hard crash if
        impl='cython_gpu', and you feed it too many bounding boxes. Ideally this will
        be fixed in the future.

        TODO: SoftNMS [SoftNMS]_.

    References:

        .. [CythonNMS] https://github.com/facebookresearch/Detectron/blob/master/detectron/utils/cython_nms.pyx

        .. [NMSPython] https://www.pyimagesearch.com/2015/02/16/faster-non-maximum-suppression-python/

        .. [SoftNMS] https://github.com/bharatsingh430/soft-nms/blob/master/lib/nms/cpu_nms.pyx

    SeeAlso:
        :func:`kwimage.structs.detections.Detections.non_max_supression`.

    CommandLine:
        xdoctest -m ~/code/kwimage/kwimage/algo/algo_nms.py non_max_supression

    Example:
        >>> from kwimage.algo.algo_nms import *
        >>> from kwimage.algo.algo_nms import _impls
        >>> ltrb = np.array([
        >>>     [0, 0, 100, 100],
        >>>     [100, 100, 10, 10],
        >>>     [10, 10, 100, 100],
        >>>     [50, 50, 100, 100],
        >>> ], dtype=np.float32)
        >>> scores = np.array([.1, .5, .9, .1])
        >>> keep = non_max_supression(ltrb, scores, thresh=0.5, impl='numpy')
        >>> print('keep = {!r}'.format(keep))
        >>> assert keep == [2, 1, 3]
        >>> thresh = 0.0
        >>> non_max_supression(ltrb, scores, thresh, impl='numpy')
        >>> if 'numpy' in available_nms_impls():
        >>>     keep = non_max_supression(ltrb, scores, thresh, impl='numpy')
        >>>     assert list(keep) == [2, 1]
        >>> if 'cython_cpu' in available_nms_impls():
        >>>     keep = non_max_supression(ltrb, scores, thresh, impl='cython_cpu')
        >>>     assert list(keep) == [2, 1]
        >>> if 'cython_gpu' in available_nms_impls():
        >>>     keep = non_max_supression(ltrb, scores, thresh, impl='cython_gpu')
        >>>     assert list(keep) == [2, 1]
        >>> if 'torch' in available_nms_impls():
        >>>     keep = non_max_supression(ltrb, scores, thresh, impl='torch')
        >>>     assert set(keep.tolist()) == {2, 1}
        >>> if 'torchvision' in available_nms_impls():
        >>>     keep = non_max_supression(ltrb, scores, thresh, impl='torchvision')  # note torchvision has no bias
        >>>     assert list(keep) == [2]
        >>> thresh = 1.0
        >>> if 'numpy' in available_nms_impls():
        >>>     keep = non_max_supression(ltrb, scores, thresh, impl='numpy')
        >>>     assert list(keep) == [2, 1, 3, 0]
        >>> if 'cython_cpu' in available_nms_impls():
        >>>     keep = non_max_supression(ltrb, scores, thresh, impl='cython_cpu')
        >>>     assert list(keep) == [2, 1, 3, 0]
        >>> if 'cython_gpu' in available_nms_impls():
        >>>     keep = non_max_supression(ltrb, scores, thresh, impl='cython_gpu')
        >>>     assert list(keep) == [2, 1, 3, 0]
        >>> if 'torch' in available_nms_impls():
        >>>     keep = non_max_supression(ltrb, scores, thresh, impl='torch')
        >>>     assert set(keep.tolist()) == {2, 1, 3, 0}
        >>> if 'torchvision' in available_nms_impls():
        >>>     keep = non_max_supression(ltrb, scores, thresh, impl='torchvision')  # note torchvision has no bias
        >>>     assert set(kwarray.ArrayAPI.tolist(keep)) == {2, 1, 3, 0}

    Example:
        >>> import ubelt as ub
        >>> ltrb = np.array([
        >>>     [0, 0, 100, 100],
        >>>     [100, 100, 10, 10],
        >>>     [10, 10, 100, 100],
        >>>     [50, 50, 100, 100],
        >>>     [100, 100, 150, 101],
        >>>     [120, 100, 180, 101],
        >>>     [150, 100, 200, 101],
        >>> ], dtype=np.float32)
        >>> scores = np.linspace(0, 1, len(ltrb))
        >>> thresh = .2
        >>> solutions = {}
        >>> if not _impls._funcs:
        >>>     _impls._lazy_init()
        >>> for impl in _impls._funcs:
        >>>     keep = non_max_supression(ltrb, scores, thresh, impl=impl)
        >>>     solutions[impl] = sorted(keep)
        >>> assert 'numpy' in solutions
        >>> print('solutions = {}'.format(ub.urepr(solutions, nl=1)))
        >>> assert ub.allsame(solutions.values())

    CommandLine:
        xdoctest -m ~/code/kwimage/kwimage/algo/algo_nms.py non_max_supression

    Example:
        >>> import ubelt as ub
        >>> # Check that zero-area boxes are ok
        >>> ltrb = np.array([
        >>>     [0, 0, 0, 0],
        >>>     [0, 0, 0, 0],
        >>>     [10, 10, 10, 10],
        >>> ], dtype=np.float32)
        >>> scores = np.array([1, 2, 3], dtype=np.float32)
        >>> thresh = .2
        >>> solutions = {}
        >>> if not _impls._funcs:
        >>>     _impls._lazy_init()
        >>> for impl in _impls._funcs:
        >>>     keep = non_max_supression(ltrb, scores, thresh, impl=impl)
        >>>     solutions[impl] = sorted(keep)
        >>> assert 'numpy' in solutions
        >>> print('solutions = {}'.format(ub.urepr(solutions, nl=1)))
        >>> assert ub.allsame(solutions.values())
    """
    if impl == 'cpu':
        warnings.warn(
            'impl="cpu" is deprecated use impl="cython_cpu" instead',
            DeprecationWarning)
        impl = 'cython_cpu'
    elif impl == 'gpu':
        warnings.warn(
            'impl="gpu" is deprecated use impl="cython_gpu" instead',
            DeprecationWarning)
        impl = 'cython_gpu'
    elif impl == 'py':
        warnings.warn(
            'impl="py" is deprecated use impl="numpy" instead',
            DeprecationWarning)
        impl = 'numpy'

    if not _impls._funcs:
        _impls._lazy_init()

    torch = sys.modules.get('torch', None)

    if ltrb.shape[0] == 0:
        return []

    if impl == 'auto':
        is_tensor = torch is not None and torch.is_tensor(ltrb)
        num = len(ltrb)
        if is_tensor:
            if ltrb.device.type == 'cuda':
                code = 'tensor0'
            else:
                code = 'tensor'
        else:
            code = 'ndarray'
        valid = _impls._valid
        impl = _heuristic_auto_nms_impl(code, num, valid)
        # print('impl._valid = {!r}'.format(_impls._valid))
        # print('impl = {!r}'.format(impl))

    elif ub.iterable(impl):
        # if impl is iterable, it is a preference order
        found = False
        for item in impl:
            if item in _impls._funcs:
                impl = item
                found = True
                break
        if not found:
            raise KeyError('Unknown impls={}'.format(impl))

    if classes is not None:
        keep = []
        for idxs in ub.group_items(range(len(classes)), classes).values():
            # cls_ltrb = ltrb.take(idxs, axis=0)
            # cls_scores = scores.take(idxs, axis=0)
            cls_ltrb = ltrb[idxs]
            cls_scores = scores[idxs]
            cls_keep = non_max_supression(cls_ltrb, cls_scores, thresh=thresh,
                                          bias=bias, impl=impl)
            keep.extend(list(ub.take(idxs, cls_keep)))
        return keep
    else:

        if impl == 'numpy':
            api = kwarray.ArrayAPI.coerce(ltrb)
            ltrb = api.numpy(ltrb)
            scores = api.numpy(scores)
            func = _impls._funcs['numpy']
            keep = func(ltrb, scores, thresh, bias=float(bias))
        elif impl == 'torch' or impl == 'torchvision':
            api = kwarray.ArrayAPI.coerce(ltrb)
            ltrb = api.tensor(ltrb).float()
            scores = api.tensor(scores).float()
            # Default output of torch impl is a mask
            if impl == 'torchvision':
                # if bias != 1:
                #     warnings.warn('torchvision only supports bias==1')
                func = _impls._funcs['torchvision']
                # Torchvision returns indices
                keep = func(ltrb, scores, iou_threshold=thresh)
            else:
                func = _impls._funcs['torch']
                flags = func(ltrb, scores, thresh=thresh, bias=float(bias))
                keep = torch.nonzero(flags).view(-1)

            # Ensure than input type is the same as output type
            keep = api.numpy(keep)
        else:
            # TODO: it would be nice to be able to pass torch tensors here
            nms = _impls._funcs[impl]
            ltrb = kwarray.ArrayAPI.numpy(ltrb)
            scores = kwarray.ArrayAPI.numpy(scores)
            ltrb = ltrb.astype(np.float32)
            scores = scores.astype(np.float32)
            if impl == 'cython_gpu':
                # TODO: if the data is already on a torch GPU can we just
                # use it?
                # HACK: we should parameterize which device is used
                if device_id is None:
                    device_id = torch.cuda.current_device()
                keep = nms(ltrb, scores, float(thresh), bias=float(bias),
                           device_id=device_id)
            elif impl == 'cython_cpu':
                keep = nms(ltrb, scores, float(thresh), bias=float(bias))
            else:
                raise KeyError(impl)
        return keep


# TODO: add soft nms bindings


if __name__ == '__main__':
    """
    CommandLine:
        xdoctest -m kwimage.algo.algo_nms
    """
    import xdoctest
    xdoctest.doctest_module(__file__)



"""
A class to make it easier to work with single colors.
"""

import numpy as np
import ubelt as ub
from . import im_core
from . import _im_color_data

__all__ = ['Color']

__todo__ = """

    - [ ] This class badly needs a rewrite to conform to the fast init / easy
          coerce constructor paradigm.

    - [ ] Make init faster an coerce more general

    - [ ] The init should take a color tuple / list / array, unmodified
          as well requiring a format argument, which might be inferred by
          coerce.

    - [ ] Should have operations to convert to/from different internal formats.

    - [ ] Keep old auto-coercing init, but have a way to disable it by default.

    - [ ] Remove dependency on colormath

"""

BASE_COLORS = _im_color_data.BASE_COLORS
TABLEAU_COLORS = _im_color_data.TABLEAU_COLORS
XKCD_COLORS = _im_color_data.XKCD_COLORS
CSS4_COLORS = _im_color_data.CSS4_COLORS
KITWARE_COLORS = _im_color_data.KITWARE_COLORS


def _lookup_colorspace_object(space):
    from colormath import color_objects
    if space == 'rgb':
        cls = color_objects.AdobeRGBColor
    elif space == 'lab':
        cls = color_objects.LabColor
    elif space == 'hsv':
        cls = color_objects.HSVColor
    elif space == 'luv':
        cls = color_objects.LuvColor
    elif space == 'cmyk':
        cls = color_objects.CMYKColor
    elif space == 'cmy':
        cls = color_objects.CMYColor
    elif space == 'xyz':
        cls = color_objects.XYZColor
    else:
        raise KeyError(space)
    return cls


def _colormath_convert(src_color, src_space, dst_space):
    """
    Uses colormath to convert colors

    Example:
        >>> # xdoctest: +REQUIRES(module:colormath)
        >>> import kwimage
        >>> from kwimage.im_color import _colormath_convert
        >>> src_color = kwimage.Color('turquoise').as01()
        >>> print('src_color = {}'.format(ub.urepr(src_color, nl=0, precision=2)))
        >>> src_space = 'rgb'
        >>> dst_space = 'lab'
        >>> lab_color = _colormath_convert(src_color, src_space, dst_space)
        ...
        >>> print('lab_color = {}'.format(ub.urepr(lab_color, nl=0, precision=2)))
        lab_color = (78.11, -70.09, -9.33)
        >>> rgb_color = _colormath_convert(lab_color, 'lab', 'rgb')
        >>> print('rgb_color = {}'.format(ub.urepr(rgb_color, nl=0, precision=2)))
        rgb_color = (0.29, 0.88, 0.81)
        >>> hsv_color = _colormath_convert(lab_color, 'lab', 'hsv')
        >>> print('hsv_color = {}'.format(ub.urepr(hsv_color, nl=0, precision=2)))
        hsv_color = (175.39, 1.00, 0.88)
    """
    from colormath.color_conversions import convert_color
    src_cls = _lookup_colorspace_object(src_space)
    dst_cls = _lookup_colorspace_object(dst_space)
    src = src_cls(*src_color)
    dst = convert_color(src, dst_cls)
    dst_color = dst.get_value_tuple()
    return dst_color


class Color(ub.NiceRepr):
    """
    Used for converting a single color between spaces and encodings.
    This should only be used when handling small numbers of colors(e.g. 1),
    don't use this to represent an image.

    CommandLine:
        xdoctest -m kwimage.im_color Color

    Example:
        >>> from kwimage import Color
        >>> print(Color('g'))
        >>> print(Color('orangered'))
        >>> print(Color('#AAAAAA').as255())
        >>> print(Color([0, 255, 0]))
        >>> print(Color([1, 1, 1.]))
        >>> print(Color([1, 1, 1]))
        >>> print(Color(Color([1, 1, 1])).as255())
        >>> print(Color(Color([1., 0, 1, 0])).ashex())
        >>> print(Color([1, 1, 1], alpha=255))
        >>> print(Color([1, 1, 1], alpha=255, space='lab'))
    """
    def __init__(self, color, alpha=None, space=None, coerce=True):
        """
        Args:
            color (Color | Iterable[int | float] | str):
                something coercable into a color
            alpha (float | None):
                if psecified adds an alpha value
            space (str):
                The colorspace to interpret this color as. Defaults to rgb.
            coerce (bool):
                The exsting init is not lightweight. This is a design problem
                that will need to be fixed in future versions. Setting
                coerce=False will disable all magic and use imputed color and
                space args directly. Alpha will be ignored.
        """
        if coerce:
            try:
                # Hack for ipython reload
                is_color_cls = color.__class__.__name__ == 'Color'
            except Exception:
                is_color_cls = isinstance(color, Color)

            if is_color_cls:
                assert alpha is None
                assert space is None
                space = color.space
                color = color.color01
            else:
                # FIXME: This is a bad check, and it's hard to fix given that there
                # is lots of code that likely depends on this now. We should not be
                # doing coercion in an `__init__`, which should alway be
                # lightweight.  The check_inputs=False can be used to disable this
                # explicitly as a workaround.
                color = self._ensure_color01(color)
                if alpha is not None:
                    alpha = self._ensure_color01([alpha])[0]

            if space is None:
                space = 'rgb'

            # always normalize the color down to 01
            color01 = list(color)

            if alpha is not None:
                if len(color01) not in [1, 3]:
                    raise ValueError('alpha already in color')
                color01 = color01 + [alpha]

            # correct space if alpha is given
            if len(color01) in [2, 4]:
                if not space.endswith('a'):
                    space += 'a'
        else:
            color01 = color
            space = space

        # FIXME: color01 is not a good name because the data wont be between 0
        # and 1 for non-rgb spaces. We should differentiate between rgb01 and
        # rgb255.
        self.color01 = color01
        self.space = space

    @classmethod
    def coerce(cls, data, **kwargs):
        return cls(data, **kwargs)

    def __nice__(self):
        colorpart = ', '.join(['{:.2f}'.format(c) for c in self.color01])
        return self.space + ': ' + colorpart

    def forimage(self, image, space='auto'):
        """
        Return a numeric value for this color that can be used
        in the given image.

        Create a numeric color tuple that agrees with the format of the input
        image (i.e. float or int, with 3 or 4 channels).

        Args:
            image (ndarray): image to return color for
            space (str): colorspace of the input image.
                Defaults to 'auto', which will choose rgb or rgba

        Returns:
            Tuple[Number, ...]: the color value

        Example:
            >>> import kwimage
            >>> img_f3 = np.zeros([8, 8, 3], dtype=np.float32)
            >>> img_u3 = np.zeros([8, 8, 3], dtype=np.uint8)
            >>> img_f4 = np.zeros([8, 8, 4], dtype=np.float32)
            >>> img_u4 = np.zeros([8, 8, 4], dtype=np.uint8)
            >>> kwimage.Color('red').forimage(img_f3)
            (1.0, 0.0, 0.0)
            >>> kwimage.Color('red').forimage(img_f4)
            (1.0, 0.0, 0.0, 1.0)
            >>> kwimage.Color('red').forimage(img_u3)
            (255, 0, 0)
            >>> kwimage.Color('red').forimage(img_u4)
            (255, 0, 0, 255)
            >>> kwimage.Color('red', alpha=0.5).forimage(img_f4)
            (1.0, 0.0, 0.0, 0.5)
            >>> kwimage.Color('red', alpha=0.5).forimage(img_u4)
            (255, 0, 0, 127)
            >>> kwimage.Color('red').forimage(np.uint8)
            (255, 0, 0)
        """
        if space == 'auto':
            space = 'rgb'
        try:
            kind = image.dtype.kind
        except AttributeError:
            kind = np.dtype(image).kind
            if len(self.color01) == 4:
                if not space.endswith('a'):
                    space = space + 'a'
        else:
            if im_core.num_channels(image) == 4:
                if not space.endswith('a'):
                    space = space + 'a'

        if kind == 'f':
            color = self.as01(space)
        else:
            color = self.as255(space)
        return color

    def _forimage(self, image, space='rgb'):
        """ backwards compat, deprecate """
        ub.schedule_deprecation(
            'kwimage', 'Color._forimage', 'method',
            migration='Use forimage instead',
            deprecate='0.10.0', error='1.0.0', remove='1.1.0')
        return self.forimage(image, space)

    def ashex(self, space=None):
        """
        Convert to hex values

        Args:
            space (None | str):
                if specified convert to this colorspace before returning

        Returns:
            str: the hex representation
        """
        c255 = self.as255(space)
        return '#' + ''.join(['{:02x}'.format(c) for c in c255])

    def as255(self, space=None):
        """
        Convert to byte values

        Args:
            space (None | str):
                if specified convert to this colorspace before returning

        Returns:
            Tuple[int, int, int] | Tuple[int, int, int, int]:
                The uint8 tuple of color values between 0 and 255.
        """
        # TODO: be more efficient about not changing to 01 space
        color = tuple(int(c * 255) for c in self.as01(space))
        return color

    def as01(self, space=None):
        """
        Convert to float values

        Args:
            space (None | str):
                if specified convert to this colorspace before returning

        Returns:
            Tuple[float, float, float] | Tuple[float, float, float, float]:
                The float tuple of color values between 0 and 1

        Note:
            This function is only guarenteed to return 0-1 values for rgb
            values. For HSV and LAB, the native spaces are used. This is not
            ideal, and we may create a new function that fixes this - at least
            conceptually - and deprate this for that in the future.

            For HSV, H is between 0 and 360. S, and V are in [0, 1]
        """
        color = tuple(map(float, self.color01))
        if space is not None:
            if space == self.space:
                pass
            elif space == 'rgb' and self.space == 'rgba':
                color = color[0:3]
            elif space == 'rgba' and self.space == 'rgb':
                color = color + (1.0,)
            elif space == 'bgr' and self.space == 'rgb':
                color = color[::-1]
            elif space == 'rgb' and self.space == 'bgr':
                color = color[::-1]
            elif space == 'lab' and self.space == 'rgb':
                # Note: in this case we will not get a 0-1 normalized color.
                # because lab does not natively exist in the 0-1 space.
                color = _colormath_convert(color, 'rgb', 'lab')
            elif space == 'hsv' and self.space == 'rgb':
                color = _colormath_convert(color, 'rgb', 'hsv')
            elif space == 'hsv' and self.space == 'rgba':
                rgba = color
                color = _colormath_convert(rgba[0:3], 'rgb', 'hsv') + rgba[3:4]
            elif space == 'rgb' and self.space == 'hsv':
                color = _colormath_convert(color, 'hsv', 'rgb')
            elif space == 'rgb' and self.space == 'lab':
                color = _colormath_convert(color, 'lab', 'rgb')
            else:
                # from colormath import color_conversions
                raise NotImplementedError('{} -> {}'.format(self.space, space))
        return color

    @classmethod
    def _is_base01(channels):
        """ check if a color is in base 01 """
        def _test_base01(channels):
            tests01 = {
                'is_float': all([isinstance(c, (float, np.float64)) for c in channels]),
                'is_01': all([c >= 0.0 and c <= 1.0 for c in channels]),
            }
            return tests01
        if isinstance(channels, str):
            return False
        return all(_test_base01(channels).values())

    @classmethod
    def _is_base255(Color, channels):
        """ there is a one corner case where all pixels are 1 or less """
        if (all(c > 0.0 and c <= 255.0 for c in channels) and any(c > 1.0 for c in channels)):
            # Definately in 255 space
            return True
        else:
            # might be in 01 or 255
            return all(isinstance(c, int) for c in channels)

    @classmethod
    def _hex_to_01(Color, hex_color):
        """
        hex_color = '#6A5AFFAF'
        """
        assert hex_color.startswith('#'), 'not a hex string %r' % (hex_color,)
        parts = hex_color[1:].strip()
        color255 = tuple(int(parts[i: i + 2], 16) for i in range(0, len(parts), 2))
        assert len(color255) in [3, 4], 'must be length 3 or 4'
        return Color._255_to_01(color255)

    def _ensure_color01(Color, color):
        """ Infer what type color is and normalize to 01 """
        if isinstance(color, str):
            color = Color._string_to_01(color)
        elif Color._is_base255(color):
            color = Color._255_to_01(color)
        return color

    @classmethod
    def _255_to_01(Color, color255):
        """ converts base 255 color to base 01 color """
        return [channel / 255.0 for channel in color255]

    @classmethod
    def _string_to_01(Color, color):
        """
        Ignore:
            mplutil.Color._string_to_01('green')
            mplutil.Color._string_to_01('red')
        """
        if color == 'random':
            import random
            ub.schedule_deprecation(
                'kwimage', 'Color._string_to_01 with random', 'method',
                migration='Use Color.random instead',
                deprecate='0.10.0', error='1.0.0', remove='1.1.0')
            color = random.choice(Color.named_colors())
        if color in BASE_COLORS:
            color01 = BASE_COLORS[color]
        elif color in CSS4_COLORS:
            color_hex = CSS4_COLORS[color]
            color01 = Color._hex_to_01(color_hex)
        elif color in XKCD_COLORS:
            color_hex = XKCD_COLORS[color]
            color01 = Color._hex_to_01(color_hex)
        elif color in KITWARE_COLORS:
            color_hex = KITWARE_COLORS[color]
            color01 = Color._hex_to_01(color_hex)
        elif color.startswith('#'):
            color01 = Color._hex_to_01(color)
        else:
            raise ValueError('unknown color=%r' % (color,))
        return color01

    @classmethod
    def named_colors(cls):
        """
        Returns:
            List[str]: names of colors that Color accepts

        Example:
            >>> import kwimage
            >>> named_colors = kwimage.Color.named_colors()
            >>> color_lut = {name: kwimage.Color(name).as01() for name in named_colors}
            >>> # xdoctest: +REQUIRES(module:kwplot)
            >>> # xdoctest: +REQUIRES(--show)
            >>> import kwplot
            >>> kwplot.autompl()
            >>> # This is a very big table if we let it be, reduce it
            >>> color_lut =dict(list(color_lut.items())[0:10])
            >>> canvas = kwplot.make_legend_img(color_lut)
            >>> kwplot.imshow(canvas)
        """
        NAMED_COLORS = set(BASE_COLORS) | set(CSS4_COLORS) | set(XKCD_COLORS) | set(KITWARE_COLORS)
        names = sorted(NAMED_COLORS)
        return names

    @classmethod
    def distinct(Color, num, existing=None, space='rgb', legacy='auto',
                 exclude_black=True, exclude_white=True):
        """
        Make multiple distinct colors.

        The legacy variant is based on a stack overflow post [HowToDistinct]_,
        but the modern variant is based on the :mod:`distinctipy` package.

        References:
            .. [HowToDistinct] https://stackoverflow.com/questions/470690/how-to-automatically-generate-n-distinct-colors

            .. [ColorLimits] https://graphicdesign.stackexchange.com/questions/3682/where-can-i-find-a-large-palette-set-of-contrasting-colors-for-coloring-many-d
            .. [WikiDistinguish] https://en.wikipedia.org/wiki/Help:Distinguishable_colors
            .. [Disinct2] https://ux.stackexchange.com/questions/17964/how-many-visually-distinct-colors-can-accurately-be-associated-with-a-separate

        TODO:
            - [ ] If num is more than a threshold we should switch to
               a different strategy to generating colors that just samples
               uniformly from some colormap and then shuffles. We have no hope
               of making things distinguishable when num starts going over
               10 or so. See [ColorLimits]_ [WikiDistinguish]_ [Disinct2]_ for
               more ideas.

        Returns:
            List[Tuple]: list of distinct float color values

        Example:
            >>> # xdoctest: +REQUIRES(module:matplotlib)
            >>> from kwimage.im_color import *  # NOQA
            >>> import kwimage
            >>> colors1 = kwimage.Color.distinct(5, legacy=False)
            >>> colors2 = kwimage.Color.distinct(3, existing=colors1)
            >>> # xdoctest: +REQUIRES(module:kwplot)
            >>> # xdoctest: +REQUIRES(--show)
            >>> from kwimage.im_color import _draw_color_swatch
            >>> swatch1 = _draw_color_swatch(colors1, cellshape=9)
            >>> swatch2 = _draw_color_swatch(colors1 + colors2, cellshape=9)
            >>> import kwplot
            >>> kwplot.autompl()
            >>> kwplot.imshow(swatch1, pnum=(1, 2, 1), fnum=1)
            >>> kwplot.imshow(swatch2, pnum=(1, 2, 2), fnum=1)
            >>> kwplot.show_if_requested()
        """
        if legacy == 'auto':
            legacy = (existing is None)

        if legacy:
            import matplotlib as mpl
            import matplotlib._cm  as _cm
            assert existing is None
            # Old behavior
            cm = mpl.colors.LinearSegmentedColormap.from_list(
                'gist_rainbow', _cm.datad['gist_rainbow'],
                mpl.rcParams['image.lut'])

            distinct_colors = [
                np.array(cm(i / num)).tolist()[0:3]
                for i in range(num)
            ]
            if space == 'rgb':
                return distinct_colors
            else:
                return [Color(c, space='rgb').as01(space=space) for c in distinct_colors]
        else:
            from distinctipy import distinctipy
            if space != 'rgb':
                raise NotImplementedError
            exclude_colors = existing
            if exclude_colors is None:
                exclude_colors = []
            if exclude_black:
                exclude_colors = exclude_colors + [(0., 0., 0.)]
            if exclude_white:
                exclude_colors = exclude_colors + [(1., 1., 1.)]
            # convert string to int for seed
            seed = int(ub.hash_data(exclude_colors, base=10)) + num
            distinct_colors = distinctipy.get_colors(
                num, exclude_colors=exclude_colors, rng=seed)
            distinct_colors = [tuple(map(float, c)) for c in distinct_colors]
            return distinct_colors

        if space == 'rgb':
            return distinct_colors
        else:
            return [Color(c, space='rgb').as01(space=space) for c in distinct_colors]

    @classmethod
    def random(Color, pool='named', with_alpha=0, rng=None):
        """
        Returns:
            Color

        Example:
            >>> import kwimage
            >>> kwimage.Color.random(pool='rgb-uniform')
            >>> kwimage.Color.random(pool='named')
        """
        import kwarray
        rng = kwarray.ensure_rng(rng, api='python')
        if pool == 'named':
            color_name = rng.choice(Color.named_colors())
            color = Color._string_to_01(color_name)
        elif pool == 'rgb-uniform':
            color = [rng.random() for _ in range(3)]
        else:
            raise NotImplementedError
        if with_alpha:
            color = color + [rng.random()]
        return Color(color)

    def distance(self, other, space='lab'):
        """
        Distance between self an another color

        Args:
            other (Color): the color to compare
            space (str): the colorspace to compare in

        Returns:
            float

        Ignore:
            import kwimage
            self = kwimage.Color((0.16304347826086973, 0.0, 1.0))
            other = kwimage.Color('purple')

            hard_coded_colors = {
                'a': (1.0, 0.0, 0.16),
                'b': (1.0, 0.918918918918919, 0.0),
                'c': (0.0, 1.0, 0.0),
                'd': (0.0, 0.9239130434782604, 1.0),
                'e': (0.16304347826086973, 0.0, 1.0)
            }

            # Find grays
            names = kwimage.Color.named_colors()
            grays = {}
            for name in names:
                color = kwimage.Color(name)
                r, g, b = color.as01()
                if r == g and g == b:
                    grays[name] = (r, g, b)
            print(ub.urepr(ub.sorted_vals(grays), nl=-1))

            for k, v in hard_coded_colors.items():
                self = kwimage.Color(v)
                distances = []
                for name in names:
                    other = kwimage.Color(name)
                    dist = self.distance(other)
                    distances.append(dist)

                idxs = ub.argsort(distances)[0:5]
                dists = list(ub.take(distances, idxs))
                names = list(ub.take(names, idxs))
                print('k = {!r}'.format(k))
                print('names = {!r}'.format(names))
                print('dists = {!r}'.format(dists))
        """
        vec1 = np.array(self.as01(space))
        vec2 = np.array(other.as01(space))
        return np.linalg.norm(vec1 - vec2)

    def nearest_named(self, with_delta=0, space='lab'):
        """
        Find the distance to the nearest named color.

        Args:
            space (str): the colorspace to compare in

        Example:
            >>> # xdoctest: +REQUIRES(module:colormath)
            >>> import kwimage
            >>> self = kwimage.Color.random(pool='rgb-uniform')
            >>> name, delta = self.nearest_named(with_delta=1)
            >>> print(f'name={name!r}, delta={delta!r}')
            >>> kwimage.Color.random().nearest_named()
        """
        # TODO: efficient lookup
        names = self.__class__.named_colors()
        named_color = [self.__class__.coerce(n) for n in names]
        distances = [self.distance(other, space=space)
                     for other in named_color]
        idx = np.argmin(distances)
        color = named_color[idx]
        name = names[idx]
        delta = np.array(self.color01) - np.array(color.color01)
        if with_delta:
            return name, delta
        else:
            return name

    def interpolate(self, other, alpha=0.5, ispace=None, ospace=None):
        """
        Interpolate between colors

        Args:
            other (Color): A coercable Color
            alpha (float | List[float]): one or more interpolation values
            ispace (str | None): colorspace to interpolate in
            ospace (str | None): colorspace of returned color

        Returns:
            Color | List[Color]

        Example:
            >>> import kwimage
            >>> color1 = self = kwimage.Color.coerce('orangered')
            >>> color2 = other = kwimage.Color.coerce('dodgerblue')
            >>> alpha = np.linspace(0, 1, 6)
            >>> ispace = 'rgb'
            >>> ospace = 'rgb'
            >>> colorBs = self.interpolate(other, alpha, ispace=ispace, ospace=ospace)
            >>> # xdoctest: +REQUIRES(module:kwplot)
            >>> # xdoctest: +REQUIRES(--show)
            >>> from kwimage.im_color import _draw_color_swatch
            >>> swatch_colors = [color1] + colorBs + [color2]
            >>> print('swatch_colors = {}'.format(ub.urepr(swatch_colors, nl=1)))
            >>> swatch1 = _draw_color_swatch(swatch_colors, cellshape=(8, 8))
            >>> import kwplot
            >>> kwplot.autompl()
            >>> kwplot.imshow(swatch1, pnum=(1, 1, 1), fnum=1)
            >>> kwplot.show_if_requested()
        """
        import kwimage
        other = kwimage.Color.coerce(other)
        vec1 = np.array(self.as01(ispace))
        vec2 = np.array(other.as01(ispace))
        if ub.iterable(alpha):
            alpha = np.asarray(alpha).ravel()
            vecB = vec1[None, :] * (1 - alpha)[:, None] + (vec2[None, :] * alpha[:, None])
            new = [
                kwimage.Color(
                    kwimage.Color(c, space=ispace, coerce=False).as01(ospace),
                    space=ospace, coerce=False)
                for c in vecB
            ]
        else:
            vecB = vec1 * (1 - alpha) + (vec2 * alpha)
            c = vecB
            new = kwimage.Color(
                kwimage.Color(c, space=ispace, coerce=False).as01(ospace),
                space=ospace, coerce=False)
        return new

    def to_image(self, dsize=(8, 8)):
        """
        Create an solid-color image with this color

        Args:
            dsize (Tuple[int, int]):
                the desired width / height of the image (defaults to 8x8)
        """
        w, h = dsize
        color_arr = np.array(self.as01()).astype(np.float32)
        cell_pixel = color_arr[None, None]
        cell = np.tile(cell_pixel, (h, w, 1))
        return cell

    def adjust(self, saturate=0, lighten=0, opacity=0):
        """
        Adjust the saturation or value of a color.

        Requires that :mod:`colormath` is installed.

        Args:
            saturate (float):
                between +1 and -1, when positive saturates the color, when
                negative desaturates the color.

            lighten (float):
                between +1 and -1, when positive lightens the color, when
                negative darkens the color

            opacity (float):
                between +1 and -1, when positive increases opacity, when
                negative decreases opacity.

        Returns:
            Self - modified color

        SeeAlso:
            :func:`kwimage.adjust` - note this variant of this has considerably
            different behavior which we may rectify in the future.

        Example:
            >>> # xdoctest: +REQUIRES(module:cv2)
            >>> # xdoctest: +REQUIRES(module:colormath)
            >>> import kwimage
            >>> self = kwimage.Color.coerce('salmon')
            >>> new = self.adjust(saturate=+0.2)
            >>> cell1 = self.to_image()
            >>> cell2 = new.to_image()
            >>> # xdoctest: +REQUIRES(--show)
            >>> import kwplot
            >>> kwplot.autompl()
            >>> canvas = kwimage.stack_images([cell1, cell2], axis=1)
            >>> kwplot.imshow(canvas)

        Example:
            >>> # xdoctest: +REQUIRES(module:cv2)
            >>> # xdoctest: +REQUIRES(module:colormath)
            >>> import kwimage
            >>> self = kwimage.Color.coerce('salmon', alpha=0.5)
            >>> new = self.adjust(saturate=+0.2)
            >>> cell1 = self.to_image()
            >>> cell2 = new.to_image()
            >>> # xdoctest: +REQUIRES(--show)
            >>> import kwplot
            >>> kwplot.autompl()
            >>> canvas = kwimage.stack_images([cell1, cell2], axis=1)
            >>> kwplot.imshow(canvas)

        Example:
            >>> # xdoctest: +REQUIRES(module:cv2)
            >>> # xdoctest: +REQUIRES(module:colormath)
            >>> import kwimage
            >>> adjustments = [
            >>>     {'saturate': -0.2},
            >>>     {'saturate': +0.2},
            >>>     {'lighten': +0.2},
            >>>     {'lighten': -0.2},
            >>>     {'saturate': -0.9},
            >>>     {'saturate': +0.9},
            >>>     {'lighten': +0.9},
            >>>     {'lighten': -0.9},
            >>>     {'opacity': +0.5},
            >>>     {'opacity': -0.5},
            >>> ]
            >>> self = kwimage.Color.coerce('kitware_green')
            >>> dsize = (256, 64)
            >>> to_show = []
            >>> to_show.append(self.to_image(dsize))
            >>> for kwargs in adjustments:
            >>>     new = self.adjust(**kwargs)
            >>>     cell = new.to_image(dsize=dsize)
            >>>     text = ub.urepr(kwargs, compact=1, nobr=1)
            >>>     cell, info = kwimage.draw_text_on_image(cell, text, return_info=1, border={'thickness': 2}, color='white', fontScale=1.0)
            >>>     to_show.append(cell)
            >>> # xdoctest: +REQUIRES(--show)
            >>> # xdoctest: +REQUIRES(module:kwplot)
            >>> import kwplot
            >>> kwplot.autompl()
            >>> canvas = kwimage.stack_images_grid(to_show)
            >>> canvas = kwimage.draw_header_text(canvas, 'kwimage.Color.adjust')
            >>> kwplot.imshow(canvas)
        """
        # TODO: work in hls instead? But don't break backwards compat?
        h, s, v, *a = self.as01(space='hsv')
        assert 0 <= h <= 360
        assert 0 <= s <= 1
        assert 0 <= v <= 1
        if saturate:
            s = max(min(1, s + saturate), 0)
        if lighten:
            v = max(min(1, v + lighten), 0)
        hsv = (h, s, v)
        rgb = _colormath_convert(hsv, 'hsv', 'rgb')

        if opacity:
            if len(a) == 0:
                a = [1]
            else:
                assert len(a) == 1
            alpha = a[0]
            a = [max(min(1, alpha + opacity), 0)]

        color = list(rgb) + a
        new_color = self.__class__.coerce(color)
        return new_color


def _draw_color_swatch(colors, cellshape=9):
    """
    Draw colors in a grid

    Ignore:
        # https://seaborn.pydata.org/tutorial/color_palettes.html

        import kwplot
        sns = kwplot.sns
        from kwimage.im_color import *  # NOQA
        from kwimage.im_color import _draw_color_swatch
        colors = sns.palettes.color_palette('deep', n_colors=10)
        swatch = _draw_color_swatch(colors)
        kwplot.imshow(swatch)
    """
    import kwimage
    import math
    if not ub.iterable(cellshape):
        cellshape = [cellshape, cellshape]
    cell_h = cellshape[0]
    cell_w = cellshape[1]
    cells = []
    for color in colors:
        cell = kwimage.Color(color).to_image(dsize=(cell_w, cell_h))
        cells.append(cell)

    num_colors = len(colors)
    num_cells_side0 = max(1, int(np.sqrt(num_colors)))
    num_cells_side1 = math.ceil(num_colors / num_cells_side0)
    num_cells = num_cells_side1 * num_cells_side0
    num_null_cells = num_cells - num_colors
    if num_null_cells > 0:
        null_cell = np.zeros((cell_h, cell_w, 3), dtype=np.float32)
        pts1 = np.array([(0, 0),                   (cell_w - 1, 0)])
        pts2 = np.array([(cell_w - 1, cell_h - 1), (0, cell_h - 1)])
        null_cell = kwimage.draw_line_segments_on_image(
            null_cell, pts1, pts2, color='red')
        # null_cell = kwimage.draw_text_on_image(
        #     {'width': cell_w, 'height': cell_h}, text='X', color='red',
        #     halign='center', valign='center')
        null_cell = kwimage.ensure_float01(null_cell)
        cells.extend([null_cell] * num_null_cells)
    swatch = kwimage.stack_images_grid(
        cells, chunksize=num_cells_side0, axis=0)
    return swatch



from numpy import ndarray
from os import PathLike
from _typeshed import Incomplete

JPG_EXTENSIONS: Incomplete
GDAL_EXTENSIONS: Incomplete
ITK_EXTENSIONS: Incomplete
IMAGE_EXTENSIONS: Incomplete


def imread(fpath: str,
           space: str = 'auto',
           backend: str = 'auto',
           **kw) -> ndarray:
    ...


def imwrite(fpath: PathLike,
            image: ndarray,
            space: str | None = 'auto',
            backend: str = 'auto',
            **kwargs) -> str:
    ...


def load_image_shape(fpath: str,
                     backend: str = 'auto',
                     include_channels: bool = True):
    ...
